#!/usr/bin/env python3

from .imageset import ImageSet

import torchvision.datasets as datasets
import torchvision.transforms as transforms
from typing import Union


class MNIST(ImageSet):

    name: str = 'mnist'
    num_classes: int = 10
    data_shape = [1, 28, 28]

    def __init__(self, norm_par={'mean': [0.1307], 'std': [0.3081]}, **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        datasets.MNIST(root=self.folder_path, train=True, download=True)
        datasets.MNIST(root=self.folder_path, train=False, download=True)

    def get_org_dataset(self, mode, transform: Union[str, object] = 'default', **kwargs):
        if transform == 'default':
            transform = self.get_transform(mode=mode)
        assert mode in ['train', 'valid']
        return datasets.MNIST(root=self.folder_path, train=(mode == 'train'), transform=transform, **kwargs)



#!/usr/bin/env python3

from trojanzoo.defenses import Defense
from trojanvision.datasets import ImageSet
from trojanvision.models import ImageModel
from trojanvision.attacks.backdoor import BadNet

import argparse


class BackdoorDefense(Defense):

    name: str = None

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--original', dest='original', action='store_true',
                           help='load original clean model, defaults to False.')

    def __init__(self, original: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.dataset: ImageSet
        self.model: ImageModel
        self.attack: BadNet  # for linting purpose
        self.original: bool = original
        self.target_class = self.attack.target_class

    def detect(self, **kwargs):
        if not self.original:
            self.attack.load(**kwargs)
        if self.attack.name == 'trojannet':
            self.model = self.attack.combined_model
        self.attack.validate_fn()

    def get_filename(self, **kwargs):
        return self.attack.name + '_' + self.attack.get_filename(**kwargs)



#!/usr/bin/env python3

from .output import prints

from typing import TYPE_CHECKING
from typing import Any, Generic, Mapping, MutableMapping, Iterator, TypeVar    # TODO: python 3.10
_KT = TypeVar("_KT")  # Key type.
_VT = TypeVar("_VT")  # Value type.
_T_co = TypeVar("_T_co", covariant=True)  # Any type covariant containers.
if TYPE_CHECKING:
    pass


class Module(MutableMapping[_KT, _VT], Generic[_KT, _VT]):
    _marker = 'M'

    def __init__(self, *args: list[Mapping[_KT, _VT]], **kwargs):
        self.__data: dict[_KT, _VT] = {}
        if len(args) == 1 and args[0] is None:
            return
        self.update(*args, **kwargs)

    def update(self, *args: list[Mapping[_KT, _VT]], **kwargs):
        args = list(args)
        args.append(kwargs)
        for module in args:
            self._update(module)
        return self

    def _update(self, module: Mapping[_KT, _VT]):    # TODO: Union[dict, Module]
        for key, value in module.items():
            if value is None:
                continue
            if key in self.keys() and isinstance(self[key], Module):
                self[key].update(value)     # TODO: linting problem
            elif key in self.keys() and isinstance(value, Module):
                self[key] = type(value)(self[key])
                self[key].update(value)
            elif isinstance(value, Module):
                self[key] = value.copy()
            else:
                self[key] = value
        return self

    def remove_none(self):
        """Remove the parameters whose values are ``None``

        :return: ``self``
        :rtype: Module
        """
        for key in list(self.__data.keys()):
            if self.__data[key] is None:
                del self.__data[key]
        return self

    def copy(self):
        return type(self)(self)

    def clear(self):
        for item in list(self.keys()):
            delattr(self, item)
        return self

    def keys(self):
        return self.__data.keys()

    def items(self):
        return self.__data.items()

    def __getattr__(self, name: _KT) -> _VT:
        if name == '_Module__data':
            super().__getattr__(name)
        return self.__data[name]

    def __getitem__(self, k: _KT):
        return self.__data[k]

    def __setattr__(self, name: _KT, value: _VT):
        if name == '_Module__data':
            return super().__setattr__(name, value)
        self.__data[name] = value

    def __setitem__(self, k: _KT, v: _VT):
        self.__data[k] = v

    def __delattr__(self, name: _KT):
        del self.__data[name]

    def __delitem__(self, v: _KT):
        del self.__data[v]

    def __str__(self):
        return self._marker + self.__data.__str__()

    def __repr__(self):
        return self._marker + self.__data.__repr__()

    def __len__(self) -> int:
        return self.__data.__len__()

    def __iter__(self) -> Iterator[_T_co]:
        return self.__data.__iter__()

    def summary(self, indent: int = 0):
        prints(self, indent=indent)


class Param(Module, Generic[_KT, _VT]):
    _marker = 'P'

    def update(self, *args: list[dict], **kwargs):
        if len(kwargs) == 0 and len(args) == 1 and not isinstance(args[0], dict) \
                and not isinstance(args[0], Module):
            self.default = args[0]
            return self
        return super().update(*args, **kwargs)

    def _update(self, module: dict[str, Any]):
        for key, value in module.items():
            if key == 'default':
                self.default = value
        super()._update(module)
        return self     # For linting purpose

    def remove_none(self):
        """Remove the parameters whose values are ``None``

        :return: ``self``
        :rtype: Module
        """
        for key in list(self.__data.keys()):
            if self.__data[key] is None and not (isinstance(key, str) and key == 'default'):
                del self.__data[key]
        return self

    def __getattr__(self, name: str) -> _VT:
        try:
            return super().__getattr__(name)
        except KeyError:    # TODO: or AttributeError better?
            return self['default']

    def __getitem__(self, key: str) -> _VT:
        if key not in self.keys():
            key = 'default'
            if 'default' not in self.keys():
                print(self)
                raise KeyError(key)
        return super().__getitem__(key)



#!/usr/bin/env python3

from .environ import env
from .output import ansi
from .tensor import to_list

import torch
import os
import tqdm
import tarfile
import zipfile

from typing import TYPE_CHECKING
from typing import Union    # TODO: python 3.10
if TYPE_CHECKING:
    import torch.utils.data


# def untar(file_path: str, target_path: str):
#     if not os.path.exists(target_path):
#         os.makedirs(target_path)
#     tar = tarfile.open(file_path)
#     names = tar.getnames()
#     if env['tqdm']:
#         names = tqdm(names)
#     for name in names:
#         tar.extract(name, path=target_path)
#     if env['tqdm']:
#         print('{upline}{clear_line}'.format(**ansi))
#     tar.close()


# def unzip(file_path: str, target_path: str):
#     with zipfile.ZipFile(file_path) as zf:
#         zf.extractall(target_path)


# def uncompress(file_path: str, target_path: str, verbose: bool = True):
#     if not os.path.exists(target_path):
#         os.makedirs(target_path)
#     if verbose:
#         print('{yellow}Uncompress file{reset}: '.format(**ansi), file_path)
#     ext = os.path.splitext(file_path)[1]
#     if ext in ['.zip']:
#         unzip(file_path, target_path)
#     elif ext in ['.tar', '.gz']:
#         untar(file_path, target_path)
#     else:
#         raise NotImplementedError(f'{file_path=}')
#     if verbose:
#         print('{green}Uncompress finished{reset}: '.format(**ansi),
#               f'{target_path}')
#         print()


class TensorListDataset(torch.utils.data.Dataset):
    def __init__(self, data: torch.Tensor = None, targets: list[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.data = data
        self.targets = to_list(targets)
        assert len(self.data) == len(self.targets)

    def __getitem__(self, index: Union[int, slice]) -> tuple[torch.Tensor, int]:
        return self.data[index], self.targets[index]

    def __len__(self):
        return len(self.targets)


class IndexDataset(TensorListDataset):
    def __getitem__(self, index: Union[int, slice]) -> tuple[torch.Tensor, int]:
        return index, self.data[index], self.targets[index]


def dataset_to_list(dataset: torch.utils.data.Dataset, label_only: bool = False) -> tuple[list, list[int]]:
    if label_only and 'targets' in dataset.__dict__.keys():
        return None, dataset.targets
    if 'data' in dataset.__dict__.keys() and 'targets' in dataset.__dict__.keys():
        return dataset.data, dataset.targets
    data, targets = zip(*dataset)
    if label_only:
        data = None
    else:
        data = list(data)
    targets = list(targets)
    return data, targets


def sample_batch(dataset: torch.utils.data.Dataset, batch_size: int = None,
                 idx: list[int] = None) -> tuple[list, list[int]]:
    if idx is None:
        assert len(dataset) >= batch_size
        idx = torch.randperm(len(dataset))[:batch_size]
    else:
        assert len(dataset) > max(idx)
    subset = torch.utils.data.Subset(dataset, idx)
    return dataset_to_list(subset)



#!/usr/bin/env python3

from ..backdoor_defense import BackdoorDefense
from trojanzoo.utils import AverageMeter, jaccard_idx
from trojanzoo.utils import to_tensor, to_numpy

import torch
import numpy as np
from sklearn import metrics
from sklearn.cluster import KMeans
import argparse


class NEO(BackdoorDefense):
    name: str = 'neo'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--seed_num', dest='seed_num', type=int,
                           help='ABS seed number, defaults to -5.')
        group.add_argument('--max_troj_size', dest='max_troj_size', type=int,
                           help='ABS max trojan trigger size (pixel number), defaults to 64.')
        group.add_argument('--remask_epoch', dest='remask_epoch', type=int,
                           help='ABS optimizing epoch, defaults to 1000.')
        group.add_argument('--remask_lr', dest='remask_lr', type=float,
                           help='ABS optimization learning rate, defaults to 0.1.')
        group.add_argument('--remask_weight', dest='remask_weight', type=float,
                           help='ABS optimization remask loss weight, defaults to 0.1.')

    def __init__(self, threshold_t: float = 80.0, k_means_num: int = 3, sample_num: int = 100, **kwargs):
        super().__init__(**kwargs)
        self.param_list['neo'] = ['threshold_t', 'k_means_num', 'sample_num']
        self.size: list[int] = [self.attack.mark.mark_height, self.attack.mark.mark_width]
        self.threshold_t = threshold_t
        self.sample_num = sample_num
        self.k_means_num = k_means_num

    def detect(self, **kwargs):
        super().detect(**kwargs)
        if not self.attack.mark.random_pos:
            self.real_mask = self.attack.mark.mask
        loader = self.dataset.get_dataloader('valid', drop_last=True, batch_size=100)
        _input, _ = next(iter(loader))
        poison_input = self.attack.add_mark(_input)
        poison_result_list, poison_mask_list = self.trigger_detect(poison_input)
        clean_result_list, clean_mask_list = self.trigger_detect(_input)
        y_true = torch.cat((torch.zeros(len(_input)), torch.ones(len(_input))))
        y_pred = torch.cat((clean_result_list, poison_result_list))
        print(y_pred)
        print("f1_score:", metrics.f1_score(y_true, y_pred))
        print("precision_score:", metrics.precision_score(y_true, y_pred))
        print("recall_score:", metrics.recall_score(y_true, y_pred))
        print("accuracy_score:", metrics.accuracy_score(y_true, y_pred))

    def trigger_detect(self, _input: torch.Tensor):
        """
        Args:
            _input (torch.Tensor): (N, C, H, W)

        """
        # get dominant color
        dom_c_list = []
        for img in _input:
            dom_c: torch.Tensor = self.get_dominant_colour(img)  # (C)
            dom_c_list.append(dom_c)
        dom_c = torch.stack(dom_c_list).unsqueeze(-1).unsqueeze(-1)  # (N, C, 1, 1)

        # generate random numbers
        height, width = _input.shape[-2:]
        pos_height: torch.Tensor = torch.randint(
            low=0, high=height - self.size[0], size=[len(_input), self.sample_num])  # (N, sample_num)
        pos_width: torch.Tensor = torch.randint(
            low=0, high=width - self.size[1], size=[len(_input), self.sample_num])  # (N, sample_num)
        pos_list: torch.Tensor = torch.stack([pos_height, pos_width]).transpose(0, -1)    # (N, sample_num, 2)
        # block potential triggers on _input
        block_input = _input.unsqueeze(1).repeat(1, self.sample_num, 1, 1, 1)  # (N, sample_num, C, H, W)
        for i in range(len(_input)):
            for j in range(self.sample_num):
                x = pos_list[i][j][0]
                y = pos_list[i][j][1]
                block_input[i, j, :, x:x + self.size[0], y:y + self.size[1]] = dom_c[i]
        # get potential triggers
        _input = to_tensor(_input)
        block_input = to_tensor(block_input)
        org_class = self.model.get_class(_input).unsqueeze(1).repeat(1, self.sample_num)   # (N, sample_num)
        block_class_list = []
        for i in range(self.sample_num):
            block_class = self.model.get_class(block_input[:, i])   # (N, sample_num)
            block_class_list.append(block_class)
        block_class = torch.stack(block_class_list, dim=1)
        potential_idx: torch.Tensor = org_class.eq(block_class).detach().cpu()   # (N, sample_num)

        # confirm triggers
        result_list = torch.zeros(len(_input), dtype=torch.bool)
        mask_shape = [_input.shape[0], _input.shape[-2], _input.shape[-1]]
        mask_list = torch.zeros(mask_shape, dtype=torch.float)  # (N, C, height, width)
        mark_class = self.attack.mark
        for i in range(len(_input)):
            print(f'input {i:3d}')
            pos_pairs = pos_list[i][~potential_idx[i]]   # (*, 2)
            if len(pos_pairs) == 0:
                continue
            for j, pos in enumerate(pos_pairs):
                self.attack.mark.height_offset = pos[0]
                self.attack.mark.width_offset = pos[1]
                mark_class.org_mark = _input[i, :, pos[0]:pos[0] + self.size[0], pos[1]:pos[1] + self.size[1]]
                mark_class.org_mask = torch.ones(self.size, dtype=torch.bool)
                mark_class.org_alpha_mask = torch.ones(self.size, dtype=torch.float)
                mark_class.mark, mark_class.mask, mark_class.alpha_mask = mark_class.mask_mark(
                    height_offset=pos[0], width_offset=pos[1])
                target_acc = self.confirm_backdoor()
                output_str = f'    {j:3d}  Acc: {target_acc:5.2f}'
                if not self.attack.mark.random_pos:
                    overlap = jaccard_idx(mark_class.mask.detach().cpu(), self.real_mask.detach().cpu(),
                                          select_num=self.size[0] * self.size[1])
                    output_str += f'  Jaccard Idx: {overlap:5.3f}'
                print(output_str)
                if target_acc > self.threshold_t:
                    result_list[i] = True
                    mask_list[i] = mark_class.mask
        return result_list, mask_list

    def get_dominant_colour(self, img: torch.Tensor, k_means_num=None):
        """[summary]

        Args:
            img (torch.Tensor): # (C, H, W)
            k_means_num (int, optional): Defaults to 3.

        """
        if k_means_num is None:
            k_means_num = self.k_means_num
        img = to_numpy(img.transpose(0, -1).flatten(end_dim=-2))    # (*, C)
        kmeans_result = KMeans(n_clusters=k_means_num).fit(img)
        unique, counts = np.unique(kmeans_result.labels_, return_counts=True)
        center = kmeans_result.cluster_centers_[unique[np.argmax(counts)]]
        return torch.tensor(center)

    def confirm_backdoor(self):
        top1 = AverageMeter('Acc@1', ':6.2f')
        for data in self.dataset.loader['valid']:
            _input, _ = self.model.get_data(data, mode='valid')
            poison_input = self.attack.add_mark(_input)
            with torch.no_grad():
                _class = self.model.get_class(_input)
                poison_class = self.model.get_class(poison_input)
            result = ~(_class.eq(poison_class))
            acc1 = result.float().sum() / result.numel() * 100
            top1.update(acc1.item(), len(_input))
        return top1.avg



#!/usr/bin/env python3

from .imageset import ImageSet

import torchvision.datasets as datasets
import torchvision.transforms as transforms
from typing import Union


class CIFAR10(ImageSet):

    name = 'cifar10'
    num_classes = 10
    data_shape = [3, 32, 32]
    class_to_idx = {'airplane': 0, 'automobile': 1, 'bird': 2, 'cat': 3,
                    'deer': 4, 'dog': 5, 'frog': 6, 'horse': 7, 'ship': 8, 'truck': 9}

    def __init__(self, norm_par: dict[str, list[float]] = {'mean': [0.4914, 0.4822, 0.4465],
                                                           'std': [0.2023, 0.1994, 0.2010], },
                 **kwargs):
        return super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        datasets.CIFAR10(root=self.folder_path, train=True, download=True)
        datasets.CIFAR10(root=self.folder_path, train=False, download=True)

    @staticmethod
    def get_transform(mode: str) -> Union[transforms.Compose, transforms.ToTensor]:
        if mode == 'train':
            transform = transforms.Compose([
                transforms.RandomCrop((32, 32), padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor()])
        else:
            transform = transforms.ToTensor()
        return transform

    def get_org_dataset(self, mode: str, transform: Union[str, object] = 'default', **kwargs) -> datasets.CIFAR10:
        assert mode in ['train', 'valid']
        if transform == 'default':
            transform = self.get_transform(mode=mode)
        return datasets.CIFAR10(root=self.folder_path, train=(mode == 'train'), transform=transform, **kwargs)


class CIFAR100(CIFAR10):
    name = 'cifar100'
    num_classes = 100

    def initialize(self):
        datasets.CIFAR100(root=self.folder_path, train=True, download=True)
        datasets.CIFAR100(root=self.folder_path, train=False, download=True)

    def get_org_dataset(self, mode: str, transform: Union[str, object] = 'default', **kwargs) -> datasets.CIFAR100:
        assert mode in ['train', 'valid']
        if transform == 'default':
            transform = self.get_transform(mode=mode)
        return datasets.CIFAR100(root=self.folder_path, train=(mode == 'train'), transform=transform, **kwargs)



#!/usr/bin/env python3

from trojanzoo.configs import config, Config
from trojanzoo.datasets import Dataset
from trojanzoo.models import Model
from trojanzoo.attacks import Attack
from trojanzoo.utils import get_name
from trojanzoo.utils.output import ansi
from trojanzoo.utils.process import Model_Process

import os

from typing import TYPE_CHECKING
from typing import Union    # TODO: python 3.10
import argparse    # TODO: python 3.10
if TYPE_CHECKING:
    pass


class Defense(Model_Process):

    name: str = None

    @staticmethod
    def add_argument(group: argparse._ArgumentGroup):
        group.add_argument('--defense', dest='defense_name')
        group.add_argument('--defense_dir', dest='defense_dir',
                           help='directory to contain defense results')

    def __init__(self, attack: Attack = None, **kwargs):
        super().__init__(**kwargs)
        self.attack: Attack = attack

    def detect(self, **kwargs):
        raise NotImplementedError()


def add_argument(parser: argparse.ArgumentParser, defense_name: str = None, defense: Union[str, Defense] = None,
                 class_dict: dict[str, type[Defense]] = None) -> argparse._ArgumentGroup:
    defense_name = get_name(name=defense_name, module=defense, arg_list=['--defense'])
    group = parser.add_argument_group('{yellow}defense{reset}'.format(**ansi), description=defense_name)
    DefenseType = class_dict[defense_name]
    return DefenseType.add_argument(group)     # TODO: Linting problem


def create(defense_name: str = None, defense: Union[str, Defense] = None, folder_path: str = None,
           dataset_name: str = None, dataset: Union[str, Dataset] = None,
           model_name: str = None, model: Union[str, Model] = None,
           config: Config = config, class_dict: dict[str, type[Defense]] = {}, **kwargs) -> Defense:
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    model_name = get_name(name=model_name, module=model, arg_list=['-m', '--model'])
    defense_name = get_name(name=defense_name, module=defense, arg_list=['--defense'])
    if dataset_name is None:
        dataset_name = config.get_full_config()['dataset']['default_dataset']
    general_config = config.get_config(dataset_name=dataset_name)['defense']
    specific_config = config.get_config(dataset_name=dataset_name)[defense_name]
    result = general_config._update(specific_config)._update(kwargs)    # TODO: linting issues

    DefenseType: type[Defense] = class_dict[defense_name]
    if folder_path is None:
        folder_path = result['defense_dir']
        if isinstance(dataset, Dataset):
            folder_path = os.path.join(folder_path, dataset.data_type, dataset.name)
        if model_name is not None:
            folder_path = os.path.join(folder_path, model_name)
        folder_path = os.path.join(folder_path, DefenseType.name)
    return DefenseType(name=defense_name, dataset=dataset, model=model, folder_path=folder_path, **result)



#!/usr/bin/env python3

from trojanvision.attacks import PGD
from trojanvision.environ import env
from trojanzoo.defenses import Defense
from trojanzoo.utils import repeat_to_batch, to_list, cos_sim
from trojanzoo.utils.output import prints

import torch
import torch.autograd
import torch.nn.functional as F
import argparse


class AdvMind(Defense):

    name: str = 'advmind'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--attack_adapt', dest='attack_adapt', action='store_true',
                           help='Adaptive attack to add fake queries.')
        group.add_argument('--fake_percent', dest='fake_percent', type=float,
                           help='fake query percentage.')
        group.add_argument('--dist', dest='dist', type=float,
                           help='fake query noise std.')
        group.add_argument('--defend_adapt', dest='defend_adapt', action='store_true',
                           help='Robust location M-estimator.')
        group.add_argument('--active', dest='active', action='store_true',
                           help='Proactive solicitation.')
        group.add_argument('--active_percent', dest='active_percent', type=float,
                           help='Active gradient weight.')

    def __init__(self, attack_adapt: bool = False, fake_percent: float = 0.3, dist: float = 50.0,
                 defend_adapt: bool = False, k: int = 1, b: float = 4e-3,
                 active: bool = False, active_percent: float = 0.1, **kwargs):
        super().__init__(**kwargs)
        self.attack: PGD = self.attack

        self.param_list['advmind'] = []
        if attack_adapt:
            self.param_list['advmind'].extend(['fake_percent', 'dist', 'active_percent'])
        if defend_adapt:
            self.param_list['advmind'].extend(['k', 'b'])
        if active:
            self.param_list['advmind'].extend(['active_percent'])

        self.attack_adapt: bool = attack_adapt
        self.fake_percent: float = fake_percent
        if not self.attack_adapt:
            self.fake_percent: float = 0.0
        self.dist: float = dist

        self.defend_adapt: bool = defend_adapt
        self.k: int = k
        self.b: float = b

        self.active: bool = active
        self.active_percent: float = active_percent

        # Temp Variable
        self.fake_query_num: int = int(self.attack.query_num * self.fake_percent)
        self.true_query_num: int = self.attack.query_num - self.fake_query_num
        self.attack_grad_list: list[torch.Tensor] = []

    def detect(self):
        zeros = torch.zeros(self.attack.iteration - 1)
        result_dict = {'draw': zeros.clone(), 'win': zeros.clone(), 'lose': zeros.clone()}
        counter = 0
        attack_succ_num = 0.
        detect_succ_num = 0.
        for i, data in enumerate(self.dataset.loader['test']):
            if counter >= 200:
                break
            print('img idx: ', i)
            _input, _label = self.model.remove_misclassify(data)
            if len(_label) == 0:
                print('misclassification')
                print()
                print('------------------------------------')
                continue
            target = self.model.generate_target(_input)
            result, _, _, attack_succ, detect_succ = self.inference(_input, target)
            for _iter in range(len(result)):
                result_dict[result[_iter]][_iter] += 1
            counter += 1
            if attack_succ != self.attack.iteration:
                attack_succ_num += 1
            if detect_succ != self.attack.iteration:
                detect_succ_num += 1
            attack_succ_rate = attack_succ_num / counter
            detect_succ_rate = detect_succ_num / counter
            print('draw: ', to_list(result_dict['draw']))
            print('win : ', to_list(result_dict['win']))
            print('lose: ', to_list(result_dict['lose']))
            print()
            print('total: ', counter)
            print('attack succ rate: ', attack_succ_rate)
            print('detect succ rate: ', detect_succ_rate)
            print('------------------------------------')

    def inference(self, _input: torch.Tensor, target: torch.Tensor):
        # ------------------------------- Init --------------------------------- #
        torch.manual_seed(env['seed'])
        if 'start' in self.output:
            self.attack.output_info(_input=_input, noise=torch.zeros_like(_input), target=target,
                                    loss_fn=lambda _X: self.model.loss(_X, target))
        self.attack_grad_list: list[torch.Tensor] = []
        # ------------------------ Attacker Seq -------------------------------- #
        seq = self.get_seq(_input, target)  # Attacker cluster sequences (iter, query_num+1, C, H, W)
        seq_centers, seq_bias = self.get_center_bias(seq)  # Defender cluster center estimate
        # seq_centers: (iter, 1, C, H, W)   seq_bias: (iter)
        # seq_centers = seq[:, 0]  # debug
        if 'start' in self.output:
            mean_error = (seq_centers[:, 0] - seq[:, 0]).abs().flatten(start_dim=1).max(dim=1)[0]
            print('Mean Shift Distance: '.ljust(25) +
                  f'avg {mean_error.mean():<10.5f} min {mean_error.min():<10.5f} max {mean_error.max():<10.5f}')
            print('Bias Estimation: '.ljust(25) +
                  f'avg {seq_bias.mean():<10.5f} min {seq_bias.min():<10.5f} max {seq_bias.max():<10.5f}')
        # candidate_centers = self.get_candidate_centers(seq, seq_centers, seq_bias)  # abandoned
        # candidate_centers = seq_centers
        detect_result = self.get_detect_result(seq_centers, target=target)
        attack_result = self.model(seq[:, 0].squeeze()).argmax(dim=1)

        attack_succ = self.attack.iteration
        detect_succ = self.attack.iteration

        detect_true = True
        for i in range(self.attack.iteration - 1):
            if attack_result[i] == target and \
                    attack_result[min(i + 1, self.attack.iteration - 2)] == target and \
                    attack_succ == self.attack.iteration:
                attack_succ = i
            if detect_result[i] == detect_result[min(i + 1, self.attack.iteration - 2)] and detect_succ == self.attack.iteration and detect_true:
                if detect_result[i] == target:
                    detect_succ = i
                else:
                    detect_true = False
        if 'end' in self.output:
            # print('candidate centers: ', [len(i) for i in candidate_centers])
            print('Detect Iter: ', detect_succ)
            prints(to_list(detect_result), indent=12)
            print('Attack Iter: ', attack_succ)
            prints(to_list(attack_result), indent=12)
            print()
        result = ['draw'] * (self.attack.iteration - 1)
        if attack_succ < detect_succ:
            for i in range(attack_succ, self.attack.iteration - 1):
                result[i] = 'lose'
        elif attack_succ > detect_succ:
            for i in range(detect_succ, self.attack.iteration - 1):
                result[i] = 'win'
        elif attack_succ == detect_succ:
            pass
        else:
            raise ValueError()
        return result, detect_result, attack_result, attack_succ, detect_succ

    def get_seq(self, _input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        seq = []
        X = _input.clone()
        noise = torch.zeros_like(_input)
        if isinstance(target, int):
            target = target * torch.ones(len(_input), dtype=torch.long, device=_input.device)

        def loss_fn(_X: torch.Tensor, **kwargs):
            t = target
            if len(_X) != len(target) and len(target) == 1:
                t = target * torch.ones(len(_X), dtype=torch.long, device=_X.device)
            return F.cross_entropy(self.model(_X), t, **kwargs)

        for _iter in range(self.attack.iteration):
            # Attacker generate sequence
            if self.attack.grad_method == 'hess' and _iter % self.hess_p == 0:  # TODO
                self.hess = self.calc_hess(loss_fn, X, sigma=self.attack.sigma,
                                           hess_b=self.hess_b, hess_lambda=self.hess_lambda)
                self.hess /= self.hess.norm(p=2)
            cluster = self.attack.gen_seq(X, query_num=self.true_query_num)
            # Attack Adaptive
            if self.attack_adapt:
                fake_seq = self.get_fake_seq(X=X)
                cluster = torch.cat([cluster, fake_seq])
            seq.append(cluster)
            # Defense Active
            if self.active:
                center = self.get_center(cluster)
                # center_idx = (cluster - center).flatten(start_dim=1).norm(p=2, dim=1).argmin()
                # center = cluster[center_idx]
                center.requires_grad_()
                loss = loss_fn(center)
                real_grad = torch.autograd.grad(loss, center)[0]
                center.requires_grad = False
                real_grad /= float(real_grad.abs().max())
                # real_grad.sign_()
                noise_grad = torch.zeros_like(real_grad).flatten()
                offset = (target + _iter) % self.model.num_classes
                for multiplier in range(len(noise_grad) // self.model.num_classes):
                    noise_grad[multiplier * self.model.num_classes + offset] = 1
                noise_grad = noise_grad.view(real_grad.shape)
                active_grad = self.active_percent * noise_grad + \
                    (1 - self.active_percent) * real_grad

                def active_loss(_X: torch.Tensor, **kwargs):
                    org_loss = loss_fn(center, **kwargs)
                    active_loss = ((_X - center) * active_grad).flatten(start_dim=1).sum(dim=1)
                    return org_loss + active_loss
                grad = self.attack.calc_seq(f=active_loss, seq=cluster[:self.true_query_num + 1])
            else:
                grad = self.attack.calc_seq(f=loss_fn, seq=cluster[:self.true_query_num + 1])
            grad.sign_()
            if 'middle' in self.output:
                self.attack_grad_list.append(grad.clone())

            noise = (noise - self.attack.alpha * grad).clamp(-self.attack.epsilon, self.attack.epsilon)
            X = (_input + noise).clamp(0, 1)
            noise = X - _input
        return torch.stack(seq)

    def get_fake_seq(self, X: torch.Tensor) -> torch.Tensor:
        if X.dim() == 4:
            X = X[0]
        noise = torch.normal(mean=0.0, std=1.0, size=X.shape, device=X.device)
        fake_seq = X + self.dist * self.attack.sigma * noise
        return repeat_to_batch(fake_seq, batch_size=self.fake_query_num)

    # cluster: (query_num, C, H, W)
    def get_center(self, cluster: torch.Tensor) -> torch.Tensor:
        if self.defend_adapt:
            T = cluster.median(dim=0)[0].unsqueeze(0)   # (1, C, H, W)
            S = cluster - T  # (query_num, C, H, W)
            S = S.flatten(start_dim=1)  # (query_num, C*H*W)
            S = 1.4826 * self.b * torch.median(S.norm(p=2, dim=1))   # scalar
            for i in range(self.k):
                T = T + S / 0.4132 * self.phi_log(
                    (cluster - T) / S).median(dim=0)[0].unsqueeze(0)   # (1, C, H, W)
            return T
        else:
            return cluster.mean(dim=0).unsqueeze(0)  # (1, C, H, W)

    def get_bias(self, cluster: torch.Tensor) -> float:
        T = cluster.median(dim=0)[0].unsqueeze(0)
        S = cluster - T
        S = S.flatten(start_dim=1)
        S = 1.4826 * self.b * torch.median(S.norm(p=2, dim=1))
        # B = math.sqrt(2)*torch.erfinv(torch.as_tensor([1 / (1-self.fake_percent)-1])).item()
        B = 0.0
        for i in range(self.k):
            result = float(self.phi_log((cluster - B) / S).median(dim=0)[0].norm(p=2))
            B = B + S / 0.4132 * (self.fake_percent + (1 - self.fake_percent) * result)
        return B

    def get_center_bias(self, seq: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        seq_centers = []
        seq_bias = []
        for cluster in seq:
            T = self.get_center(cluster)
            seq_centers.append(T)
            B = self.get_bias(cluster)
            seq_bias.append(B)
        return torch.stack(seq_centers), torch.stack(seq_bias)

    def get_detect_result(self, seq_centers: torch.Tensor, target=None):
        pair_seq = -torch.ones(self.attack.iteration - 1, dtype=torch.long, device=env['device'])
        detect_prob = torch.ones(self.model.num_classes) / self.model.num_classes
        for i in range(len(seq_centers) - 1):
            X: torch.Tensor = seq_centers[i].clone()
            dist_list = torch.zeros(self.model.num_classes)

            for _class in range(self.model.num_classes):
                _label = _class * torch.ones(len(X), dtype=torch.long, device=X.device)
                X.requires_grad_()
                loss = self.model.loss(X, _label)
                grad = torch.autograd.grad(loss, X)[0]
                X.requires_grad = False
                grad /= grad.abs().max()
                if self.active:
                    noise_grad = torch.zeros(X.numel(), device=X.device)
                    offset = (_class + i) % self.model.num_classes

                    for multiplier in range(len(noise_grad) // self.model.num_classes):
                        noise_grad[multiplier * self.model.num_classes + offset] = 1
                    noise_grad = noise_grad.view(X.shape)
                    # noise_grad /= noise_grad.abs().max()
                    grad = self.active_percent * noise_grad + \
                        (1 - self.active_percent) * grad
                grad.sign_()
                vec = seq_centers[i + 1] - X
                dist = cos_sim(-grad, vec)
                dist_list[_class] = dist
                if 'middle' in self.output and _class == target:
                    print('sim <vec, real>: ', cos_sim(vec, -grad))
                    print('sim <est, real>: ',
                          cos_sim(self.attack_grad_list[i], grad))
                    print('sim <vec, est>: ',
                          cos_sim(vec, -self.attack_grad_list[i]))
            # todo: Use atanh for normalization after pytorch 1.6
            detect_prob = F.softmax(torch.log((2 / (1 - dist_list)).sub(1)))
            # detect_prob.div_(detect_prob.norm(p=2))
            pair_seq[i] = detect_prob.argmax().item()
        return pair_seq

    @staticmethod
    def phi_log(x: torch.Tensor) -> torch.Tensor:
        return 1 - 2 / (x.exp() + 1)

    # -------------------------- Unused ----------------------------- #

    @staticmethod
    def phi(x: torch.Tensor, c: float = 1.35) -> torch.Tensor:
        return x.clamp(-c, c)

    # Unused TODO
    @staticmethod
    def get_candidate_centers(seq: torch.Tensor, seq_centers: torch.Tensor, seq_bias: torch.Tensor):
        center_seq = []
        for i in range(len(seq)):
            # for point in seq[i]:
            #     sub_seq.append(point)

            # sub_seq: list[torch.Tensor] = []
            # idx = float(to_tensor([(x-seq_centers[i]).norm(p=2)
            #                  for x in seq[i]]).argmin())
            # sub_seq.append(seq[i][idx])
            # norms = [(x - seq_centers[i]).norm(p=2) for x in sub_seq]
            norms: torch.Tensor = (seq - seq_centers[i]).flatten(1).norm(p=2, dim=-1)
            if norms.dim():
                idx = norms.argmin()
                center_seq.append([seq[idx]])
            else:
                center_seq.append([])
        return center_seq

    # Unused
    def get_center_class_pairs(self, candidate_centers: torch.Tensor, seq_centers: torch.Tensor, seq: torch.Tensor):
        pair_seq = []
        for i in range(len(candidate_centers) - 1):
            sub_pair_seq = []
            for point in candidate_centers[i]:
                # if self.active:
                #     vec = seq_centers[i+1]-point
                #     _result = vec.view(-1)
                #     for j in range(len(_result)):
                #         if _result[j] < 0 and j > i:
                #             sub_pair_seq.append((j-i) % self.num_classes)
                # print(vec.view(-1)[:self.num_classes])
                x = point.clone()
                dist_list = torch.zeros(self.model.num_classes)
                # print('bound: ', estimate_error + shift_dist)
                for _class in range(self.model.num_classes):
                    x.requires_grad_()
                    loss = self.model.loss(x, _class)
                    grad = torch.autograd.grad(loss, x)[0]
                    x.requires_grad = False
                    grad.sign_()
                    if self.active:
                        noise_grad = torch.zeros_like(grad).flatten()
                        offset = (_class + i) % self.model.num_classes
                        for multiplier in range(int(len(noise_grad) / self.model.num_classes)):
                            noise_grad[multiplier * self.model.num_classes + offset] = 1
                        noise_grad = noise_grad.view(grad.shape)
                        grad = self.active_percent * noise_grad + \
                            (1 - self.active_percent) * grad
                        grad.sign_()
                    vec = seq_centers[i + 1] - point
                    dist = cos_sim(-grad, vec)
                    dist_list[_class] = dist
                sub_pair_seq.append(dist_list.argmax().item())
            pair_seq.append(sub_pair_seq)
        return pair_seq



#!/usr/bin/env python3

from ..backdoor_defense import BackdoorDefense
from trojanvision.environ import env
from trojanzoo.utils import to_tensor, normalize_mad, jaccard_idx
from trojanzoo.utils import AverageMeter
from trojanzoo.utils import onehot_label
from trojanzoo.utils.output import prints, ansi, output_iter

import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import time
import datetime
import os
import argparse
from tqdm import tqdm


mse_criterion = nn.MSELoss()


class DeepInspect(BackdoorDefense):

    name: str = 'deep_inspect'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--sample_ratio', dest='sample_ratio', type=float,
                           help='sample ratio from the full training data')
        group.add_argument('--noise_dim', dest='noise_dim', type=int,
                           help='GAN noise dimension')
        group.add_argument('--remask_epoch', dest='remask_epoch', type=int,
                           help='Remask optimizing epoch')
        group.add_argument('--remask_lr', dest='remask_lr', type=float,
                           help='Remask optimizing learning rate')
        group.add_argument('--gamma_1', dest='gamma_1', type=float,
                           help='control effect of GAN loss')
        group.add_argument('--gamma_2', dest='gamma_2', type=float,
                           help='control effect of perturbation loss')

    def __init__(self, sample_ratio: float = 0.1, noise_dim: int = 100,
                 remask_epoch: int = 30, remask_lr=0.01,
                 gamma_1: float = 0.0, gamma_2: float = 1, **kwargs):
        super().__init__(**kwargs)

        self.param_list['deep_inspect'] = ['sample_ratio', 'remask_epoch', 'remask_lr', 'gamma_1', 'gamma_2']

        self.sample_ratio: int = sample_ratio
        self.remask_epoch: int = remask_epoch
        self.remask_lr: float = remask_lr
        self.gamma_1: float = gamma_1
        self.gamma_2: float = gamma_2

        self.noise_dim: int = noise_dim

        dataset = self.dataset.get_dataset(mode='train')
        subset, _ = self.dataset.split_set(dataset, percent=sample_ratio)
        self.loader = self.dataset.get_dataloader(mode='train', dataset=subset)

    def detect(self, **kwargs):
        super().detect(**kwargs)
        if not self.attack.mark.random_pos:
            self.real_mask = self.attack.mark.mask
        loss_list, mark_list = self.get_potential_triggers()
        np.savez(os.path.join(self.folder_path, self.get_filename(target_class=self.target_class) + '.npz'),
                 mark_list=mark_list, loss_list=loss_list)
        print('loss: ', loss_list)
        print('loss MAD: ', normalize_mad(loss_list))

    def get_potential_triggers(self) -> tuple[torch.Tensor, torch.Tensor]:
        mark_list, loss_list = [], []
        # todo: parallel to avoid for loop
        for label in range(self.model.num_classes):
            print('Class: ', output_iter(label, self.model.num_classes))
            loss, mark = self.remask(label)
            loss_list.append(loss)
            mark_list.append(mark)
        loss_list = torch.as_tensor(loss_list)
        return loss_list, mark_list

    def load(self, path: str = None):
        if path is None:
            path = os.path.join(self.folder_path, self.get_filename() + '.npz')
        _dict = np.load(path, allow_pickle=True)
        self.attack.mark.mark = to_tensor(_dict['mark_list'][self.attack.target_class])
        self.attack.mark.random_pos = False
        self.attack.mark.height_offset = 0
        self.attack.mark.width_offset = 0

        def add_mark_fn(_input, **kwargs):
            return _input + self.attack.mark.mark.to(_input.device)
        self.attack.mark.add_mark_fn = add_mark_fn

    def remask(self, label: int) -> tuple[torch.Tensor, torch.Tensor]:
        generator = Generator(self.noise_dim, self.dataset.num_classes, self.dataset.data_shape)
        for param in generator.parameters():
            param.requires_grad_()
        optimizer = optim.Adam(generator.parameters(), lr=self.remask_lr)
        optimizer.zero_grad()
        # mask = self.attack.mark.mask

        losses = AverageMeter('Loss', ':.4e')
        entropy = AverageMeter('Entropy', ':.4e')
        norm = AverageMeter('Norm', ':.4e')
        acc = AverageMeter('Acc', ':6.2f')
        torch.manual_seed(env['seed'])
        noise = torch.rand(1, self.noise_dim, device=env['device'])
        mark = torch.zeros(self.dataset.data_shape, device=env['device'])
        for _epoch in range(self.remask_epoch):
            losses.reset()
            entropy.reset()
            norm.reset()
            acc.reset()
            epoch_start = time.perf_counter()
            loader = self.loader
            if env['tqdm']:
                loader = tqdm(loader)
            for data in loader:
                _input, _label = self.model.get_data(data)
                batch_size = _label.size(0)
                poison_label = label * torch.ones_like(_label)
                mark = generator(noise, torch.tensor([label], device=poison_label.device, dtype=poison_label.dtype))
                poison_input = (_input + mark).clamp(0, 1)
                _output = self.model(poison_input)

                batch_acc = poison_label.eq(_output.argmax(1)).float().mean()
                batch_entropy = self.model.criterion(_output, poison_label)
                batch_norm = mark.flatten(start_dim=1).norm(p=1, dim=1).mean()
                batch_loss = batch_entropy + self.gamma_2 * batch_norm

                acc.update(batch_acc.item(), batch_size)
                entropy.update(batch_entropy.item(), batch_size)
                norm.update(batch_norm.item(), batch_size)
                losses.update(batch_loss.item(), batch_size)

                batch_loss.backward()
                optimizer.step()
                optimizer.zero_grad()
            epoch_time = str(datetime.timedelta(seconds=int(
                time.perf_counter() - epoch_start)))
            pre_str = '{blue_light}Epoch: {0}{reset}'.format(
                output_iter(_epoch + 1, self.remask_epoch), **ansi).ljust(64 if env['color'] else 35)
            _str = ' '.join([
                f'Loss: {losses.avg:.4f},'.ljust(20),
                f'Acc: {acc.avg:.2f}, '.ljust(20),
                f'Norm: {norm.avg:.4f},'.ljust(20),
                f'Entropy: {entropy.avg:.4f},'.ljust(20),
                f'Time: {epoch_time},'.ljust(20),
            ])
            prints(pre_str, _str, prefix='{upline}{clear_line}'.format(**ansi) if env['tqdm'] else '', indent=4)

        def get_data_fn(data, **kwargs):
            _input, _label = self.model.get_data(data)
            poison_label = torch.ones_like(_label) * label
            poison_input = (_input + mark).clamp(0, 1)
            return poison_input, poison_label
        self.model._validate(print_prefix='Validate Trigger Tgt', get_data_fn=get_data_fn, indent=4)

        if not self.attack.mark.random_pos:
            overlap = jaccard_idx(mark.mean(dim=0), self.real_mask,
                                  select_num=self.attack.mark.mark_height * self.attack.mark.mark_width)
            print(f'    Jaccard index: {overlap:.3f}')

        for param in generator.parameters():
            param.requires_grad = False
        return losses.avg, mark

    # def loss(self, _output: torch.Tensor, mark: torch.Tensor, poison_label: torch.Tensor) -> torch.Tensor:
    #     loss_trigger = self.model.criterion(_output, poison_label)
    #     # loss_gan = self.loss_gan(_output, label)
    #     loss_pert = self.loss_pert(mark)
    #     return loss_trigger + self.gamma_2 * loss_pert  # self.gamma_1 * loss_gan +

    # def loss_gan(self, _output: torch.Tensor, label: int) -> torch.Tensor:
    #     onehot_label = torch.zeros_like(_output)
    #     onehot_label[:, label] = 1.0
    #     return self.mse_criterion(_output, onehot_label)

    # @staticmethod
    # def loss_pert(self, mark: torch.Tensor) -> torch.Tensor:
    #     return mark.norm(p=1)


class Generator(nn.Module):
    def __init__(self, noise_dim: int = 100, num_classes: int = 10, data_shape: list[int] = [3, 32, 32]):
        self.noise_dim: int = noise_dim
        self.num_classes: int = num_classes
        self.data_shape: list[int] = data_shape
        super(Generator, self).__init__()

        self.relu = nn.ReLU(inplace=True)
        self.sigmoid = nn.Sigmoid()
        self.fc2 = nn.Linear(self.num_classes, 1000)
        self.fc = nn.Linear(self.noise_dim + 1000, 64 * self.data_shape[1] * self.data_shape[2])
        self.bn1 = nn.BatchNorm2d(64)
        self.deconv1 = nn.ConvTranspose2d(64, 32, 5, 1, 2)
        self.bn2 = nn.BatchNorm2d(32)
        self.deconv2 = nn.ConvTranspose2d(32, self.data_shape[0], 5, 1, 2)
        if env['num_gpus']:
            self.cuda()

    def forward(self, noise: torch.Tensor, poison_label: torch.Tensor) -> torch.Tensor:
        _label = onehot_label(poison_label, self.num_classes).float()
        y_ = self.fc2(_label)
        y_ = self.relu(y_)
        x = torch.cat([noise, y_], dim=1)
        x: torch.Tensor = self.fc(x)
        x = x.view(-1, 64, self.data_shape[1], self.data_shape[2])
        x = self.bn1(x)
        x = self.relu(x)
        x = self.deconv1(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.deconv2(x)
        x = self.sigmoid(x)
        return x



#!/usr/bin/env python3

from ..backdoor_defense import BackdoorDefense
from trojanvision.environ import env

import torch
import torch.utils.data
from torch.utils.data import TensorDataset
import argparse


class SpectralSignature(BackdoorDefense):

    """
    Spectral Signature Defense is described in the paper 'Spectral Signatures in Backdoor Attacks'_ by Brandon Tran. The main idea is backdoor attack tends to leave behind a detectable trace in the spectrum of the covariance of a feature representation learned by the neural network, that is if the means of the two populations are sufﬁciently well-separated relative to the variance of the populations, the corrupted datapoints can be detected and removed using singular value decomposition. 
    The intuition is that if the set of inputs with a given label consists of both clean examples as well as corrupted examples from a different label set, the backdoor from the latter set will provide a strong signal in this representation for classiﬁcation. As long as the signal is large in magnitude, we can detect it via singular value decomposition and remove the images that provide the signal. 

    The authors have not posted `original source code`_. 

    Args:
        preprocess_layer (str): the chosen layer used to extract feature representation. Default: features.
        poison_image_num (int): the number of sampled poison image to train the  initial model. Default: 50.
        clean_image_num (int): the number of sampled clean image to train the  initial model. Default: 500.
        epsilon (int): the number of examples to remove from each class. Default: 5.
        retrain_epoch (int): the epoch to retrain the model on clean image dataset. Default: 10.

    .. _Spectral_Signature:
        https://arxiv.org/abs/1811.00636


    Returns:
        model(ImageModel): the clean model trained only with clean samples.
    """
    name: str = 'spectral_signature'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--preprocess_layer', dest='preprocess_layer', type=str,
                           help='the chosen layer used to extract feature representation, defaults to ``flatten``')
        group.add_argument('--poison_image_num', dest='poison_image_num', type=int,
                           help='the number of sampled poison image to train the model initially, defaults to 50')
        group.add_argument('--clean_image_num', dest='clean_image_num', type=int,
                           help='the number of sampled clean image to train the model initially, defaults to 500')
        group.add_argument('--epsilon', dest='epsilon', type=int,
                           help='the number of examples to remove from each class, defaults to 5')
        group.add_argument('--retrain_epoch', dest='retrain_epoch', type=int,
                           help='the epoch to retrain the model on clean image dataset, defaults to 5')

    def __init__(self, poison_image_num: int = 50, clean_image_num: int = 500, preprocess_layer: str = 'features', epsilon: int = 5, retrain_epoch: int = 5, **kwargs):
        super().__init__(**kwargs)
        self.preprocess_layer: str = preprocess_layer
        self.poison_image_num: int = poison_image_num
        self.clean_image_num: int = clean_image_num
        self.epsilon: int = epsilon
        self.retrain_epoch: int = retrain_epoch

        self.clean_dataset, _ = self.dataset.split_set(
            dataset=self.dataset.get_full_dataset(mode='train'), length=self.clean_image_num)
        label_all = torch.empty([])    # TODO
        clean_input_all = torch.empty([])    # TODO
        for i, data in enumerate(iter(self.clean_dataset)):
            _input, _label = self.model.get_data(data)
            clean_input = _input.view(1, _input.shape[0], _input.shape[1], _input.shape[2])
            if i == 0:
                clean_input_all = clean_input
                label_all = torch.unsqueeze(_label, 0)
            else:
                clean_input_all = torch.cat((clean_input_all, clean_input))
                label_all = torch.cat((label_all, torch.unsqueeze(_label, 0)))
        label_all = torch.squeeze(label_all, 0)
        self.clean_dataset = TensorDataset(clean_input_all, label_all)
        self.clean_dataloader = self.dataset.get_dataloader(mode='train', dataset=self.clean_dataset, num_workers=0)

        self.poison_dataset, _ = self.dataset.split_set(dataset=_, length=self.poison_image_num)
        label_all = torch.empty([])    # TODO
        poison_input_all = torch.empty([])    # TODO
        for i, data in enumerate(iter(self.poison_dataset)):
            _input, _label = self.model.get_data(data)
            poison_input = self.attack.add_mark(_input)
            poison_input = poison_input.view(1, poison_input.shape[0], poison_input.shape[1], poison_input.shape[2])
            if i == 0:
                poison_input_all = poison_input
                label_all = torch.unsqueeze(_label, 0)
            else:
                poison_input_all = torch.cat((poison_input_all, poison_input))
                label_all = torch.cat((label_all, torch.unsqueeze(_label, 0)))
        label_all = torch.squeeze(label_all, 0)
        self.poison_dataset = TensorDataset(poison_input_all, label_all)
        self.poison_dataloader = self.dataset.get_dataloader(
            mode='train', dataset=self.poison_dataset, num_workers=0, pin_memory=False)

        self.mix_dataset = torch.utils.data.ConcatDataset([self.clean_dataset, self.poison_dataset])
        self.mix_dataloader = self.dataset.get_dataloader(
            mode='train', dataset=self.mix_dataset, num_workers=0, pin_memory=False)

    def detect(self, optimizer, lr_scheduler, **kwargs):
        """
        Record the detected poison samples, remove them and retrain the model from scratch to get a clean model.
        """
        super().detect(**kwargs)
        initial_model = self.model
        self.model._train(optimizer=optimizer, lr_scheduler=lr_scheduler, loader_train=self.mix_dataloader, **kwargs)
        final_loader = self.get_clean_dataloader()
        initial_model._train(epoch=self.retrain_epoch, optimizer=optimizer,
                             lr_scheduler=lr_scheduler, loader_train=final_loader)

    def get_clean_dataloader(self):
        """
        Get the feature representation of samples from each class individually.
        Compute centered  representation.
        Singular value decomposition.
        Compute the vector of the outlier scores.
        Remove the examples with the top epsilon scores from the samples of each class and form the clean dataset.

        Returns:
            torch.utils.data.DataLoader: after removing the suspicious samples with bigger singular value, return the clean dataloader.
        """
        final_set = None    # TODO
        for k in range(self.dataset.num_classes):
            # self.class_dataset = self.dataset.get_class_set(self.mix_dataset,classes = [k])
            idx = []
            for i, data in enumerate(self.mix_dataset):
                _input, _label = self.model.get_data(data)
                _input = _input.view(1, _input.shape[0], _input.shape[1], _input.shape[2])
                if _label.item() == k:
                    idx.append(k)
            self.class_dataset = torch.utils.data.Subset(self.mix_dataset, idx)

            layer_output_all = torch.empty([])    # TODO
            for i, data in enumerate(self.class_dataset):
                _input, _label = self.model.get_data(data)
                layer_output = self.model.get_layer(_input, layer_output=self.preprocess_layer)
                layer_output = layer_output.view(1, -1)
                if i == 0:
                    layer_output_all = layer_output
                else:
                    layer_output_all = torch.cat((layer_output_all, layer_output))

            layer_output_mean = torch.mean(layer_output_all, dim=0)

            for i in range(len(self.class_dataset)):
                layer_output_all[i] = layer_output_all[i] - layer_output_mean

            u, s, v = torch.svd(layer_output_all)
            v_transpose = torch.transpose(v, 1, 0)
            outlier_scores = torch.rand([layer_output_all.shape[0]], device=env['device'])
            for i in range(len(self.class_dataset)):
                outlier_scores[i] = (layer_output_all[i].view(1, -1) @ v_transpose[i].view(-1, 1)) ** 2
            outlier_scores_sorted, indices = torch.sort(outlier_scores, descending=True)

            clean_indices = indices[self.epsilon:]
            self.class_dataset = torch.utils.data.Subset(self.class_dataset, clean_indices)
            if k == 0:
                final_set = self.class_dataset
            else:
                final_set = torch.utils.data.ConcatDataset([final_set, self.class_dataset])

        final_dataloader = self.dataset.get_dataloader(mode=None, dataset=final_set, num_workers=0, pin_memory=False)
        return final_dataloader



#!/usr/bin/env python3

from trojanvision.models import ImageModel
from trojanzoo.utils import to_list
import torch


class Curvature():
    # TODO: finish the detect()
    def __init__(self, model: ImageModel = None, h: float = 0.1, **kwargs):
        self.h: float = h
        self.model: ImageModel = model

    def compute_gradient(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        x = x.clone().detach()
        x.requires_grad = True
        loss = self.model.loss(x, y)
        grad = torch.autograd.grad(loss, x)[0]
        return grad

    def measure(self, x: torch.Tensor, y: torch.Tensor, d=None):
        gx1 = self.compute_gradient(x, y)
        if d is None:
            d = gx1.sign().flatten(start_dim=1)
            d /= torch.norm(d, dim=-1, keepdim=True)
            d = d.view_as(x)
        x2 = x + self.h * d
        gx2 = self.compute_gradient(x2, y)
        diff = (gx2 - gx1).flatten(start_dim=1)
        return (diff * diff).sum(1)

    def benign_measure(self, validloader=None, batch_num=20):
        if validloader is None:
            validloader = self.model.dataset.loader['valid']
        measure_list = []
        for i, data in enumerate(validloader):
            _input, _label = self.model.get_data(data)
            if i >= batch_num:
                break
            measure = self.measure(_input, _label)
            measure_list.extend(to_list(measure))
        return measure_list



#!/usr/bin/env python3

from .imagefolder import ImageFolder
import torchvision.transforms as transforms
import numpy as np


class GTSRB(ImageFolder):

    name = 'gtsrb'
    data_shape = [3, 32, 32]
    num_classes = 43
    valid_set = False
    url = {'train': 'https://sid.erda.dk/public/archives/daaeac0d7ce1152aea9b61d9f1e19370/GTSRB-Training_fixed.zip'}
    md5 = {'train': '513f3c79a4c5141765e10e952eaa2478'}
    org_folder_name = {'train': 'GTSRB/Training'}

    def __init__(self, norm_par: dict[str, list[float]] = {'mean': [0.3403, 0.3121, 0.3214],
                                                           'std': [0.2724, 0.2608, 0.2669], },
                 loss_weights: bool = True, **kwargs):
        return super().__init__(norm_par=norm_par, loss_weights=loss_weights, **kwargs)

    @staticmethod
    def get_transform(mode: str) -> transforms.Compose:
        transform = transforms.Compose([
            transforms.Resize((32, 32)),
            transforms.ToTensor()])
        return transform

    def initialize_npz(self, mode_list: list[str] = ['train', 'valid'],
                       transform: transforms.Compose = transforms.Compose([transforms.Resize((32, 32)),
                                                                           transforms.Lambda(lambda x: np.array(x))]),
                       **kwargs):
        super().initialize_npz(mode_list=mode_list, transform=transform, **kwargs)



#!/usr/bin/env python3

import os
import matplotlib.font_manager

dirname = os.path.dirname(__file__)

for style in ['normal', 'italic', 'bold', 'bold_italic']:
    file_path = os.path.normpath(os.path.join(dirname, f'palatino_{style}.ttf'))
    matplotlib.font_manager.fontManager.addfont(file_path)
ttflist: list[matplotlib.font_manager.FontEntry] = matplotlib.font_manager.fontManager.ttflist
for i, font in enumerate(ttflist):
    if 'Palatino.ttc' in font.fname:
        del ttflist[i]



#!/usr/bin/env python3

from .imc import IMC
from trojanvision.models import MagNet

import torch
import math
import random


class IMC_MagNet(IMC):
    name: str = 'imc_magnet'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.magnet = MagNet(dataset=self.dataset, pretrain=True)

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor], keep_org: bool = True, poison_label=True, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        decimal, integer = math.modf(self.poison_num)
        integer = int(integer)
        if random.uniform(0, 1) < decimal:
            integer += 1
        if not keep_org:
            integer = len(_label)
        if not keep_org or integer:
            org_input, org_label = _input, _label
            _input = self.add_mark(org_input[:integer])
            _input = self.magnet(_input)
            _label = _label[:integer]
            if poison_label:
                _label = self.target_class * torch.ones_like(org_label[:integer])
            if keep_org:
                _input = torch.cat((_input, org_input))
                _label = torch.cat((_label, org_label))
        return _input, _label



#!/usr/bin/env python3

import torch.nn.init as init
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.utils import _pair


def weight_init(m: nn.Module):
    # Function for Initialization
    '''
    Usage:
        model = Model()
        model.apply(weight_init)
    '''
    if hasattr(m, 'reset_parameters'):
        return m.reset_parameters()
    for layer in m.children():
        if hasattr(layer, 'reset_parameters'):
            return layer.reset_parameters()
    if isinstance(m, nn.Conv1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.BatchNorm1d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm2d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm3d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.Linear):
        init.xavier_normal_(m.weight.data)
        init.normal_(m.bias.data)
    elif isinstance(m, nn.LSTM):
        for param in m.parameters():
            if param.dim() >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.LSTMCell):
        for param in m.parameters():
            if param.dim() >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRU):
        for param in m.parameters():
            if param.dim() >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRUCell):
        for param in m.parameters():
            if param.dim() >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)


def conv2d_same_padding(input, weight, bias=None, stride=1, padding=1, dilation=1, groups=1):
    """
    Conv2d layer with padding=same
    the padding param here is not important
    """

    input_rows = input.size(2)
    filter_rows = weight.size(2)
    effective_filter_size_rows = (filter_rows - 1) * dilation[0] + 1
    out_rows = (input_rows + stride[0] - 1) // stride[0]
    padding_rows = max(0, (out_rows - 1) * stride[0]
                       + (filter_rows - 1) * dilation[0] + 1 - input_rows)
    rows_odd = (padding_rows % 2 != 0)
    padding_cols = max(0, (out_rows - 1) * stride[0]
                       + (filter_rows - 1) * dilation[0] + 1 - input_rows)
    cols_odd = (padding_rows % 2 != 0)

    if rows_odd or cols_odd:
        input = F.pad(input, [0, int(cols_odd), 0, int(rows_odd)])

    return F.conv2d(input, weight, bias, stride[0],
                    padding=(padding_rows // 2, padding_cols // 2),
                    dilation=dilation, groups=groups)


class Conv2d_SAME(_ConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros'):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super(Conv2d_SAME, self).__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            False, _pair(0), groups, bias, padding_mode)

    def forward(self, input):
        return conv2d_same_padding(input, self.weight, self.bias, self.stride,
                                   self.padding, self.dilation, self.groups)



#!/usr/bin/env python3

from .imc import IMC
from trojanvision.environ import env
from trojanzoo.utils import to_tensor
from trojanzoo.utils.output import prints

import torch
import numpy as np
import math
import random
import os
import argparse
from typing import Callable


class IMC_Adaptive(IMC):
    name: str = 'imc_adaptive'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--abs_weight', dest='abs_weight', type=float)
        group.add_argument('--strip_percent', dest='strip_percent', type=float)

    def __init__(self, seed_num: int = -5, count_mask: bool = True,
                 samp_k: int = 1, same_range: bool = False, n_samples: int = 5,
                 strip_percent: float = 1.0, abs_weight: float = 1e-5, **kwargs):
        super().__init__(**kwargs)

        self.seed_num: int = seed_num
        if self.seed_num < 0:
            self.seed_num = self.model.num_classes * abs(self.seed_num)
        self.count_mask: bool = count_mask

        # -----------Neural Sampling------------- #
        self.samp_k: int = samp_k
        self.same_range: bool = same_range
        self.n_samples: int = n_samples
        self.top_n_neurons: int = 20

        self.seed_data = self.load_seed_data()
        self.neuron_list = []

        self.strip_percent: float = strip_percent
        self.abs_weight: float = abs_weight

    def attack(self, epoch: int, **kwargs):
        super().attack(epoch, loss_fn=self.loss_fn, **kwargs)

    def epoch_fn(self, **kwargs):
        self.save()
        _input, _label = self.seed_data['input'], self.seed_data['label']
        all_ps = self.sample_neuron(_input)
        self.neuron_list: list[dict] = self.find_min_max(all_ps, _label)[0]
        self.optimize_mark()

    # ---------------------------- Seed Data --------------------------- #

    def save_seed_data(self) -> dict[str, np.ndarray]:
        torch.manual_seed(env['seed'])
        if self.seed_num % self.model.num_classes:
            raise ValueError(f'seed_num({self.seed_num:d}) % num_classes({self.model.num_classes:d}) should be 0.')
        seed_class_num: int = self.seed_num // self.model.num_classes
        x, y = [], []
        for _class in range(self.model.num_classes):
            loader = self.dataset.get_dataloader(mode='train', batch_size=seed_class_num, classes=[_class],
                                                 shuffle=True, num_workers=0, pin_memory=False)
            _input, _label = next(iter(loader))
            x.append(_input)
            y.append(_label)
        x = torch.cat(x).numpy()
        y = torch.cat(y).numpy()
        seed_data = {'input': x, 'label': y}
        seed_path = f'{env["result_dir"]}{self.dataset.name}/{self.name}_{self.seed_num}.npy'
        np.save(seed_path, seed_data)
        print('seed data saved at: ', seed_path)
        return seed_data

    def load_seed_data(self) -> dict[str, torch.Tensor]:
        seed_path = f'{env["result_dir"]}{self.dataset.name}/{self.name}_{self.seed_num}.npy'
        seed_data: dict[str, torch.Tensor] = {}
        seed_data = np.load(seed_path, allow_pickle=True).item() if os.path.exists(seed_path) \
            else self.save_seed_data()
        seed_data['input'] = to_tensor(seed_data['input'])
        seed_data['label'] = to_tensor(seed_data['label'], dtype=torch.long)
        return seed_data

    # -----------------------Neural Sample---------------------------- #

    def sample_neuron(self, _input: torch.Tensor) -> dict[str, torch.Tensor]:
        all_ps: dict[str, torch.Tensor] = {}
        batch_size = _input.shape[0]

        layer_output = self.model.get_all_layer(_input)
        for layer in self.model.get_layer_name():
            if 'pool' in layer or layer in ['features', 'flatten', 'classifier', 'logits', 'output']:
                continue
            cur_layer_output: torch.Tensor = layer_output[layer].detach().cpu()  # (batch_size, C, H, W)
            channel_num: int = cur_layer_output.shape[1]  # channels

            repeat_shape = [channel_num, self.n_samples]
            repeat_shape.extend([1] * cur_layer_output.dim())
            h_t: torch.Tensor = cur_layer_output.repeat(repeat_shape)
            # (C, n_samples, batch_size, C, H, W)

            vs = self.samp_k * torch.arange(self.n_samples, device=h_t.device, dtype=torch.float)
            if not self.same_range:
                maxes = cur_layer_output.max()
                vs *= float(maxes) / self.n_samples
            vs_shape = [1] * cur_layer_output.dim()
            vs_shape[0] = -1
            vs = vs.view(vs_shape)
            # (n_samples, 1, 1, 1)
            # todo: might use parallel to avoid for loop (torch.Tensor.scatter?)
            for neuron in range(channel_num):
                h_t[neuron, :, :, neuron] = vs
            # todo: the shape is too large
            # result = self.model.get_layer(h_t.flatten(end_dim=2), layer_input=layer).detach().cpu()
            result = []
            for h in h_t:
                h = h.to(device=env['device'])
                result.append(self.model.get_layer(h.flatten(end_dim=1), layer_input=layer).detach().cpu())
            result = torch.cat(result)

            result_shape = list(h_t.shape)[:3]
            result_shape.extend(list(result.shape)[1:])
            result = result.view(result_shape)
            all_ps[layer] = result
            # (C, n_samples, batch_size, num_classes)
        return all_ps

    def find_min_max(self, all_ps: dict[str, torch.Tensor], _label: torch.Tensor) -> dict[int, list[dict]]:
        neuron_dict: dict[int, list] = {i: [] for i in range(self.model.num_classes)}
        _label = _label.cpu()
        for layer in all_ps.keys():
            ps = all_ps[layer]  # (C, n_samples, batch_size, num_classes)
            vs: torch.Tensor = ps[:, self.n_samples // 5:].max(dim=1)[0] \
                - ps[:, :self.n_samples // 5].min(dim=1)[0]  # (C, batch_size, num_classes)
            values, labels = vs.sort(dim=-1, descending=True)
            condition1 = labels[:, :, 0].eq(_label)  # exclude the ground-truth labels
            values = torch.where(condition1, values[:, :, 1] - values[:, :, 2],
                                 values[:, :, 0] - values[:, :, 1])  # (C, batch_size)
            labels = torch.where(condition1, labels[:, :, 1], labels[:, :, 0])  # (C, batch_size)

            mode_labels = labels.mode(keepdim=True)[0]  # (C, 1) The most frequent label
            mode_idx = labels.eq(mode_labels)  # (C, batch_size)
            mode_labels_counts = mode_idx.sum(dim=-1)  # (C)
            condition2 = mode_labels_counts.ge(self.seed_num * 0.75)
            idx_list = condition2.nonzero().flatten().tolist()
            idx_list = sorted(idx_list, key=lambda idx: float(values[idx][mode_idx[idx]].min()))[:self.top_n_neurons]
            for idx in idx_list:
                value = float(values[idx][mode_idx[idx]].min())
                neuron_dict[int(mode_labels[idx])].append({'layer': layer, 'neuron': int(idx), 'value': value})
        return neuron_dict
    # -------------------------ReMask--------------------------------- #

    def abs_loss(self, layer_dict: dict[str, torch.Tensor], layer: str, neuron: int):
        feats = layer_dict[layer]
        vloss1 = feats[:, neuron].sum()
        vloss2 = feats.sum() - vloss1
        return -vloss1 + 1e-3 * vloss2

    def loss_fn(self, _input: torch.Tensor, _label: torch.Tensor, **kwargs) -> torch.Tensor:
        loss = self.model.loss(_input, _label)
        idx = 0
        for i in range(len(_label)):
            if _label[len(_label) - 1 - i] != self.target_class:
                break
            idx = len(_label) - 1 - i
        _input = _input[idx:]
        _label = _label[idx:]
        layer_dict = self.model.get_all_layer(_input)
        for sub_dict in self.neuron_list:
            layer = sub_dict['layer']
            neuron = sub_dict['neuron']
            loss -= self.abs_weight * self.abs_loss(layer_dict, layer, neuron)
        return loss

    def add_strip_mark(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        return self.mark.add_mark(x, alpha=1 - (1 - self.mark.mark_alpha) / 2, **kwargs)

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor], **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)

        decimal, integer = math.modf(self.poison_num)
        integer = int(integer)
        if random.uniform(0, 1) < decimal:
            integer += 1
        if integer:
            org_input, org_label = _input, _label
            poison_input = self.add_mark(org_input[:integer])
            poison_label = self.target_class * torch.ones_like(org_label[:integer])
            strip_num: float = integer * self.strip_percent
            strip_decimal, strip_integer = math.modf(strip_num)
            strip_integer = int(strip_integer)
            if random.uniform(0, 1) < strip_decimal:
                strip_integer += 1
            if strip_integer:
                strip_input = self.add_strip_mark(org_input[:strip_integer])
                strip_label = org_label[:strip_integer]
                _input = torch.cat((_input, strip_input))
                _label = torch.cat((_label, strip_label))
            _input = torch.cat((_input, poison_input))
            _label = torch.cat((_label, poison_label))
        return _input, _label

    def get_poison_data(self, data: tuple[torch.Tensor, torch.Tensor], poison_label: bool = True, strip: bool = False, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        integer = len(_label)
        if strip:
            _input = self.add_strip_mark(_input[:integer])
        else:
            _input = self.add_mark(_input[:integer])
        if poison_label:
            _label = self.target_class * torch.ones_like(_label[:integer])
        else:
            _label = _label[:integer]
        return _input, _label

    def validate_fn(self,
                    get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    loss_fn: Callable[..., torch.Tensor] = None,
                    main_tag: str = 'valid', indent: int = 0, **kwargs) -> tuple[float, float]:
        _, clean_acc = self.model._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                            get_data_fn=None, indent=indent, **kwargs)
        _, target_acc = self.model._validate(print_prefix='Validate Trigger Tgt', main_tag='valid trigger target',
                                             get_data_fn=self.get_poison_data, indent=indent, **kwargs)
        self.model._validate(print_prefix='Validate Trigger Org', main_tag='',
                             get_data_fn=self.get_poison_data, poison_label=False,
                             indent=indent, **kwargs)
        self.model._validate(print_prefix='Validate STRIP Tgt', main_tag='',
                             get_data_fn=self.get_poison_data, strip=True,
                             indent=indent, **kwargs)
        self.model._validate(print_prefix='Validate STRIP Org', main_tag='',
                             get_data_fn=self.get_poison_data, strip=True, poison_label=False,
                             indent=indent, **kwargs)
        prints(f'Validate Confidence: {self.validate_confidence():.3f}', indent=indent)
        prints(f'Neuron Jaccard Idx: {self.check_neuron_jaccard():.3f}', indent=indent)
        if self.clean_acc - clean_acc > 3 and self.clean_acc > 40:  # TODO: better not hardcoded
            target_acc = 0.0
        return clean_acc, target_acc

    def get_filename(self, mark_alpha: float = None, target_class: int = None, **kwargs):
        if mark_alpha is None:
            mark_alpha = self.mark.mark_alpha
        if target_class is None:
            target_class = self.target_class
        _file = '{mark}_tar{target:d}_alpha{mark_alpha:.2f}_mark({mark_height:d},{mark_width:d})'.format(
            mark=os.path.split(self.mark.mark_path)[1][:-4],
            target=target_class, mark_alpha=mark_alpha,
            mark_height=self.mark.mark_height, mark_width=self.mark.mark_width)
        if self.mark.random_pos:
            _file = 'random_pos_' + _file
        if self.mark.mark_distributed:
            _file = 'distributed_' + _file
        _file += f'strippercent_{self.strip_percent:f}_absweight_{self.abs_weight:f}'
        return _file



#!/usr/bin/env python3

from .output import ansi, prints, output_iter

import os

from typing import TYPE_CHECKING
from typing import Union    # TODO: python 3.10
from trojanzoo.datasets import Dataset    # TODO: python 3.10
from trojanzoo.models import Model
if TYPE_CHECKING:
    pass


class Process:
    name: str = 'process'

    def __init__(self, output: Union[int, list[str]] = 0, indent: int = 0, **kwargs):

        self.param_list: dict[str, list[str]] = {}
        self.param_list['verbose'] = ['output', 'indent']

        self.output: list[str] = None
        self.output = self.get_output(output)
        self.indent = indent

    # -----------------------------------Output-------------------------------------#
    def summary(self, indent: int = None):
        indent = indent if indent is not None else self.indent
        prints('{blue_light}{0:<20s}{reset} Parameters: '.format(self.name, **ansi), indent=indent)
        for key, value in self.param_list.items():
            prints('{green}{0:<20s}{reset}'.format(key, **ansi), indent=indent + 10)
            prints({v: getattr(self, v) for v in value}, indent=indent + 10)
            prints('-' * 20, indent=indent + 10)

    def get_output(self, org_output: Union[int, list[str]] = None) -> list[str]:
        output = None
        if org_output is None:
            output = self.output
        elif isinstance(org_output, list):
            output = set(org_output)
        elif isinstance(org_output, int):
            output = self.get_output_int(org_output)
        else:
            output = org_output
        return output

    @staticmethod
    def get_output_int(org_output: int = 0) -> list[str]:
        result: list[str] = []
        if org_output >= 5:
            result.append('end')
        if org_output >= 10:
            result.append('start')
        if org_output >= 20:
            result.append('middle')
        if org_output >= 30:
            result.append('memory')
        return result

    @staticmethod
    def output_iter(name: str, _iter: int, iteration: int = None, indent: int = 0):
        string = name + ' Iter: ' + output_iter(_iter + 1, iteration)
        prints(string, indent=indent)


class Model_Process(Process):

    name: str = 'model_process'

    def __init__(self, dataset: Dataset = None, model: Model = None, folder_path: str = None, **kwargs):
        super().__init__(**kwargs)
        self.param_list['process'] = ['clean_acc', 'folder_path']
        self.dataset: Dataset = dataset
        self.model: Model = model

        _, self.clean_acc = self.model._validate(print_prefix='Baseline Clean', get_data_fn=None, verbose=False)
        # ----------------------------------------------------------------------------- #
        self.folder_path = folder_path
        if folder_path is not None:
            self.folder_path = os.path.normpath(folder_path)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)



#!/usr/bin/env python3

from .latent_backdoor import LatentBackdoor


class IMC_Latent(LatentBackdoor):
    name: str = 'imc_latent'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.mark.random_pos:
            raise Exception('IMC requires "random pos" to be False to train mark.')

    def attack(self, **kwargs):
        print('Sample Data')
        self.data = self.sample_data()
        print('Retrain')
        return super(LatentBackdoor, self).attack(epoch_fn=self.epoch_fn, **kwargs)

    def epoch_fn(self, **kwargs):
        self.avg_target_feats = self.get_avg_target_feats(self.data)
        self.preprocess_mark(data=self.data)



#!/usr/bin/env python3

from .badnet import BadNet
from .trojannn import TrojanNN
from .latent_backdoor import LatentBackdoor
from .imc import IMC
from .reflection_backdoor import ReflectionBackdoor
from .bypass_embed import BypassEmbed
from .trojannet import TrojanNet
from .clean_label import CleanLabel
from .hidden_trigger import HiddenTrigger

from .term_study import TermStudy
from .unlearn import Unlearn

from .imc_latent import IMC_Latent
from .imc_advtrain import IMC_AdvTrain
from .imc_strip import IMC_STRIP
from .imc_multi import IMC_Multi
from .imc_magnet import IMC_MagNet
from .imc_abs import IMC_ABS
from .imc_adaptive import IMC_Adaptive



#!/usr/bin/env python3

from .badnet import BadNet
from trojanvision.attacks.adv import PGD    # TODO: Need to check whether this will cause ImportError
from trojanvision.environ import env
from trojanzoo.utils.output import ansi

import torch
import argparse
from tqdm import tqdm


class TrojanNN(BadNet):

    r"""
    TrojanNN Backdoor Attack is described in detail in the paper `TrojanNN`_ by Yingqi Liu. 

    Based on :class:`trojanzoo.attacks.backdoor.BadNet`,
    TrojanNN preprocesses the watermark pixel values to maximize the neuron activation on rarely used neurons
    to avoid the negative impact of model performance on clean iamges.

    The authors have posted `original source code`_.

    Args:
        preprocess_layer (str): The preprocess layer.
        threshold (float): the target class. Default: ``5``.
        target_value (float): The proportion of malicious images in the training set (Max 0.5). Default: 10.

    .. _TrojanNN:
        https://github.com/PurduePAML/TrojanNN/blob/master/trojan_nn.pdf

    .. _original source code:
        https://github.com/PurduePAML/TrojanNN
    """

    name: str = 'trojannn'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--preprocess_layer', dest='preprocess_layer', type=str,
                           help='the chosen feature layer patched by trigger where rare neuron activation is maxmized, defaults to ``flatten``')
        group.add_argument('--threshold', dest='threshold', type=float,
                           help='Trojan Net Threshold, defaults to 5')
        group.add_argument('--target_value', dest='target_value', type=float,
                           help='Trojan Net Target_Value, defaults to 10')
        group.add_argument('--neuron_lr', dest='neuron_lr', type=float,
                           help='Trojan Net learning rate in neuron preprocessing, defaults to 0.015')
        group.add_argument('--neuron_epoch', dest='neuron_epoch', type=int,
                           help='Trojan Net epoch in neuron preprocessing, defaults to 20')
        group.add_argument('--neuron_num', dest='neuron_num', type=int,
                           help='Trojan Net neuron numbers in neuron preprocessing, defaults to 2')

    def __init__(self, preprocess_layer: str = 'flatten', threshold: float = 5, target_value: float = 10,
                 neuron_lr: float = 0.015, neuron_epoch: int = 20, neuron_num: int = 2, **kwargs):
        super().__init__(**kwargs)
        if self.mark.random_pos:
            raise Exception('TrojanNN requires "random pos" to be False to max activate neurons.')

        self.param_list['trojannn'] = ['preprocess_layer', 'threshold', 'target_value',
                                       'neuron_lr', 'neuron_epoch', 'neuron_num']
        self.preprocess_layer: str = preprocess_layer
        self.threshold: float = threshold
        self.target_value: float = target_value

        self.neuron_lr: float = neuron_lr
        self.neuron_epoch: int = neuron_epoch
        self.neuron_num: int = neuron_num
        self.neuron_idx = None

        self.pgd = PGD(alpha=self.neuron_lr, epsilon=1.0, iteration=self.neuron_epoch, output=0, **kwargs)

    def attack(self, *args, **kwargs):
        self.neuron_idx = self.get_neuron_idx()
        self.mark.mark = self.preprocess_mark(mark=self.mark.mark * self.mark.mask, neuron_idx=self.neuron_idx)
        super().attack(*args, **kwargs)

    # get the neuron idx for preprocess.
    def get_neuron_idx(self) -> torch.Tensor:
        with torch.no_grad():
            result = []
            loader = self.dataset.loader['train']
            if env['tqdm']:
                loader = tqdm(loader)
            for i, data in enumerate(loader):
                _input, _label = self.model.get_data(data)
                fm = self.model.get_layer(_input, layer_output=self.preprocess_layer)
                if fm.dim() > 2:
                    fm = fm.flatten(start_dim=2).mean(dim=2)
                fm = fm.mean(dim=0)
                result.append(fm.detach())
            if env['tqdm']:
                print('{upline}{clear_line}'.format(**ansi))
            return torch.stack(result).sum(dim=0).argsort(descending=False)[:self.neuron_num]

    def get_neuron_value(self, x: torch.Tensor, neuron_idx: torch.Tensor) -> torch.Tensor:
        return self.model.get_layer(x, layer_output=self.preprocess_layer)[:, neuron_idx].mean()

    # train the mark to activate the least-used neurons.
    def preprocess_mark(self, mark: torch.Tensor, neuron_idx: torch.Tensor, **kwargs) -> torch.Tensor:
        with torch.no_grad():
            print("Neuron Value Before Preprocessing: ",
                  self.get_neuron_value(mark, neuron_idx))

        def loss_fn(X: torch.Tensor):
            fm = self.model.get_layer(X, layer_output=self.preprocess_layer)
            loss = fm[:, neuron_idx].mean(dim=0) - self.target_value
            return loss.norm(p=2)
        noise = torch.zeros_like(mark)
        x = mark
        for _iter in range(self.neuron_epoch):
            cost = loss_fn(x)
            if cost < self.threshold:
                break
            x, _ = self.pgd.craft_example(mark, noise=noise, iteration=1, loss_fn=loss_fn)
            noise = noise * self.mark.mask
            x = x * self.mark.mask
        x = x.detach()
        with torch.no_grad():
            print("Neuron Value After Preprocessing: ",
                  self.get_neuron_value(x, neuron_idx))
        return x

    def validate_fn(self, get_data_fn=None, **kwargs) -> tuple[float, float]:
        if self.neuron_idx is not None:
            with torch.no_grad():
                print("Neuron Value After Preprocessing: ",
                      self.get_neuron_value(self.mark.mark * self.mark.mask, self.neuron_idx))
        return super().validate_fn(get_data_fn=get_data_fn, **kwargs)



#!/usr/bin/env python3

from .badnet import BadNet
from trojanvision.utils.model import weight_init
from trojanvision.attacks.adv import PGD    # TODO: Need to check whether this will cause ImportError
from trojanvision.optim import PGD as PGD_Optimizer
from trojanvision.environ import env
from trojanzoo.utils.data import TensorListDataset

import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data

import argparse
import os


class CleanLabel(BadNet):
    r"""
    Contributor: Xiangshan Gao, Ren Pang

    Clean Label Backdoor Attack is described in detail in the paper `Clean Label Backdoor Attack`_ by Alexander Turner.

    The main idea is to perturb the poisoned samples in order to render learning the salient characteristic of the input more difficult,causing the model rely more heavily on the backdoor pattern in order to successfully introduce backdoor. Utilize the adversarial examples and GAB generated data, the resulted poisoned inputs appear to be consistent with their label and thus seem benign even upon human inspection.

    The authors haven't posted `original source code`_.

    Args:
        poison_generation_method (str): the chosen method to generate poisoned sample. Default: 'pgd'.
        tau (float): the interpolation constant used to balance source imgs and target imgs. Default: 0.4.
        epsilon (float): the perturbation bound in input space. Default: 0.1.
        noise_dim (int): the dimension of the input in the generator. Default: 100.
        generator_iters (int): the epoch for training the generator. Default: 1000.
        critic_iter (int): the critic iterations per generator training iteration. Default: 5.


    .. _Clean Label:
        https://people.csail.mit.edu/madry/lab/cleanlabel.pdf

    .. _related code:
        https://github.com/igul222/improved_wgan_training
        https://github.com/MadryLab/cifar10_challenge
        https://github.com/caogang/wgan-gp

    """
    name: str = 'clean_label'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--poison_generation_method', dest='poison_generation_method', type=str,
                           help='the chosen method to generate poisoned sample, defaults to config[clean_label][poison_generation_method]=pgd')
        group.add_argument('--tau', dest='tau', type=float,
                           help='the interpolation constant used to balance source imgs and target imgs, defaults to config[clean_label][tau]=0.2')
        group.add_argument('--epsilon', dest='epsilon', type=float,
                           help='the perturbation bound in input space, defaults to config[clean_label][epsilon]=0.1, 300/(3*32*32)')
        group.add_argument('--noise_dim', dest='noise_dim', type=int,
                           help='the dimension of the input in the generator, defaults to config[clean_label][noise_dim]=100')
        group.add_argument('--train_gan', dest='train_gan', action='store_true',
                           help='whether train the GAN if it already exists, defaults to False')
        group.add_argument('--generator_iters', dest='generator_iters', type=int,
                           help=' the epoch for training the generator, defaults to config[clean_label][generator_iters]=1000')
        group.add_argument('--critic_iter', dest='critic_iter', type=int,
                           help=' the critic iterations per generator training iteration, defaults to config[clean_label][critic_iter]=5')

    def __init__(self, preprocess_layer: str = 'classifier', poison_generation_method: str = 'pgd',
                 pgd_alpha: float = 2 / 255, pgd_epsilon: float = 16 / 255, pgd_iteration=20,
                 tau: float = 0.2, noise_dim: int = 100,
                 train_gan: bool = False, generator_iters: int = 1000, critic_iter: int = 5, **kwargs):
        super().__init__(**kwargs)
        self.param_list['clean_label'] = ['poison_generation_method', 'poison_num']
        self.poison_generation_method: str = poison_generation_method
        if poison_generation_method == 'pgd':
            self.poison_num: int = int(len(self.dataset.get_dataset(
                'train', classes=[self.target_class])) * self.poison_percent)
        elif poison_generation_method == 'gan':
            self.poison_num: int = int(len(self.dataset.get_dataset('train')) * self.poison_percent)

        if poison_generation_method == 'pgd':
            self.param_list['pgd'] = ['pgd_alpha', 'pgd_epsilon', 'pgd_iteration']
            self.pgd_alpha: float = pgd_alpha
            self.pgd_epsilon: float = pgd_epsilon
            self.pgd_iteration: int = pgd_iteration
            self.pgd: PGD = PGD(alpha=pgd_alpha, epsilon=pgd_epsilon, iteration=pgd_iteration,
                                target_idx=0, output=self.output, dataset=self.dataset, model=self.model)
        elif poison_generation_method == 'gan':
            self.param_list['gan'] = ['tau', 'noise_dim', 'train_gan', 'critic_iter', 'generator_iters']
            self.tau: float = tau
            self.noise_dim: int = noise_dim
            self.train_gan: bool = train_gan
            self.generator_iters = generator_iters
            self.critic_iter = critic_iter
            self.wgan = WGAN(noise_dim=self.noise_dim, dim=64, data_shape=self.dataset.data_shape,
                             generator_iters=self.generator_iters, critic_iter=self.critic_iter)

    def attack(self, optimizer: torch.optim.Optimizer, lr_scheduler: torch.optim.lr_scheduler._LRScheduler, **kwargs):

        target_class_dataset = self.dataset.get_dataset('train', full=True, classes=[self.target_class])

        sample_target_class_dataset, target_original_dataset = self.dataset.split_set(
            target_class_dataset, self.poison_num)
        sample_target_dataloader = self.dataset.get_dataloader(mode='train', dataset=sample_target_class_dataset,
                                                               batch_size=self.poison_num, num_workers=0)
        target_imgs, _ = self.model.get_data(next(iter(sample_target_dataloader)))

        full_set = self.dataset.get_dataset('train', full=False)
        poison_set: TensorListDataset = None    # TODO
        if self.poison_generation_method == 'pgd':
            poison_label = self.target_class * torch.ones(len(target_imgs), dtype=torch.long, device=target_imgs.device)

            poison_imgs, _ = self.model.remove_misclassify(data=(target_imgs, poison_label))
            poison_imgs, _ = self.pgd.craft_example(_input=poison_imgs)
            poison_imgs = self.add_mark(poison_imgs).cpu()

            poison_label = [self.target_class] * len(target_imgs)
            poison_set = TensorListDataset(poison_imgs, poison_label)
            # poison_set = torch.utils.data.ConcatDataset([poison_set, target_original_dataset])

        elif self.poison_generation_method == 'gan':
            other_classes = list(range(self.dataset.num_classes))
            other_classes.pop(self.target_class)
            x_list = []
            y_list = []
            for source_class in other_classes:
                print('Process data of Source Class: ', source_class)
                source_class_dataset = self.dataset.get_dataset(mode='train', full=True, classes=[source_class])
                sample_source_class_dataset, _ = self.dataset.split_set(
                    source_class_dataset, self.poison_num)
                sample_source_class_dataloader = self.dataset.get_dataloader(mode='train', dataset=sample_source_class_dataset,
                                                                             batch_size=self.poison_num, num_workers=0)
                source_imgs, _ = self.model.get_data(next(iter(sample_source_class_dataloader)))

                g_path = f'{self.folder_path}gan_dim{self.noise_dim}_class{source_class}_g.pth'
                d_path = f'{self.folder_path}gan_dim{self.noise_dim}_class{source_class}_d.pth'
                if os.path.exists(g_path) and os.path.exists(d_path) and not self.train_gan:
                    self.wgan.G.load_state_dict(torch.load(g_path, map_location=env['device']))
                    self.wgan.D.load_state_dict(torch.load(d_path, map_location=env['device']))
                    print(f'    load model from: \n        {g_path}\n        {d_path}', )
                else:
                    self.train_gan = True
                    self.wgan.reset_parameters()
                    gan_dataset = torch.utils.data.ConcatDataset([source_class_dataset, target_class_dataset])
                    gan_dataloader = self.dataset.get_dataloader(
                        mode='train', dataset=gan_dataset, batch_size=self.dataset.batch_size, num_workers=0)
                    self.wgan.train(gan_dataloader)
                    torch.save(self.wgan.G.state_dict(), g_path)
                    torch.save(self.wgan.D.state_dict(), d_path)
                    print(f'GAN Model Saved at : \n{g_path}\n{d_path}')
                    continue
                source_encode = self.wgan.get_encode_value(source_imgs).detach()
                target_encode = self.wgan.get_encode_value(target_imgs).detach()
                # noise = torch.randn_like(source_encode)
                # from trojanzoo.utils.tensor import save_tensor_as_img
                # source_img = self.wgan.G(source_encode)
                # target_img = self.wgan.G(target_encode)
                # for i in range(len(source_img)):
                #     save_tensor_as_img(f'./imgs/source_{i}.png', source_img[i])
                # for i in range(len(target_img)):
                #     save_tensor_as_img(f'./imgs/target_{i}.png', target_img[i])
                # exit()
                interpolation_encode = source_encode * self.tau + target_encode * (1 - self.tau)
                poison_imgs = self.wgan.G(interpolation_encode).detach()
                poison_imgs = self.add_mark(poison_imgs)

                poison_label = [self.target_class] * len(poison_imgs)
                poison_imgs = poison_imgs.cpu()
                x_list.append(poison_imgs)
                y_list.extend(poison_label)
            assert not self.train_gan
            x_list = torch.cat(x_list)
            poison_set = TensorListDataset(x_list, y_list)
            # poison_set = torch.utils.data.ConcatDataset([poison_set, target_original_dataset])
        final_set = torch.utils.data.ConcatDataset([poison_set, full_set])
        # final_set = poison_set
        final_loader = self.dataset.get_dataloader(mode='train', dataset=final_set, num_workers=0)
        self.model._train(optimizer=optimizer, lr_scheduler=lr_scheduler, save_fn=self.save,
                          loader_train=final_loader, validate_fn=self.validate_fn, **kwargs)


class Generator(nn.Module):
    def __init__(self, noise_dim: int = 100, dim: int = 64, data_shape: list[int] = [3, 32, 32]):
        super().__init__()
        self.noise_dim: int = noise_dim
        self.dim: int = dim
        self.data_shape: list[int] = data_shape
        init_dim = dim * data_shape[1] * data_shape[2] // 16
        self.preprocess = nn.Linear(noise_dim, init_dim)
        self.preprocess_1 = nn.Sequential(
            nn.BatchNorm2d(init_dim),
            nn.ReLU(True),)
        self.block1 = nn.Sequential(
            nn.ConvTranspose2d(4 * dim, 2 * dim, 2, stride=2),
            nn.BatchNorm2d(2 * dim),
            nn.ReLU(True),)
        self.block2 = nn.Sequential(
            nn.ConvTranspose2d(2 * dim, dim, 2, stride=2),
            nn.BatchNorm2d(dim),
            nn.ReLU(True),)
        self.deconv_out = nn.ConvTranspose2d(dim, data_shape[0], 2, stride=2)
        self.tanh = nn.Tanh()

    def forward(self, x: torch.Tensor):
        # (N, noise_dim)
        x = self.preprocess(x)
        # (N, noise_dim)
        x = x.unsqueeze(-1).unsqueeze(-1)
        x = self.preprocess_1(x)
        x = x.view(len(x), 4 * self.dim, self.data_shape[1] // 8, self.data_shape[2] // 8)
        x = self.block1(x)
        x = self.block2(x)
        x = self.deconv_out(x)
        x = self.tanh(x)
        return x


class Discriminator(nn.Module):
    def __init__(self, dim: int = 64, data_shape: list = [3, 32, 32]):
        super(Discriminator, self).__init__()
        self.dim = dim
        self.main = nn.Sequential(
            nn.Conv2d(data_shape[0], dim, 3, 2, padding=1),
            nn.LeakyReLU(),
            nn.Conv2d(dim, 2 * dim, 3, 2, padding=1),
            nn.BatchNorm2d(2 * dim),
            nn.LeakyReLU(),
            nn.Conv2d(2 * dim, 4 * dim, 3, 2, padding=1),
            nn.BatchNorm2d(4 * dim),
            nn.LeakyReLU(),
        )
        init_dim = dim * data_shape[1] * data_shape[2] // 16
        self.linear = nn.Linear(init_dim, 1)

    def forward(self, x: torch.Tensor):
        x = self.main(x)
        x = x.flatten(start_dim=1)
        x = self.linear(x)
        return x


class WGAN(object):
    def __init__(self, noise_dim: int, dim: int, data_shape: list[int] = [3, 32, 32],
                 generator_iters: int = 1000, critic_iter: int = 5):
        self.noise_dim = noise_dim
        self.G: Generator = Generator(noise_dim, dim, data_shape)
        self.D: Discriminator = Discriminator(dim, data_shape)
        if env['num_gpus']:
            self.G.cuda()
            self.D.cuda()
        # the parameter in the original paper
        self.d_optimizer = optim.RMSprop(self.D.parameters(), lr=5e-5)
        self.g_optimizer = optim.RMSprop(self.G.parameters(), lr=5e-5)
        self.generator_iters = generator_iters  # larger: 1000
        self.critic_iter = critic_iter
        self.mse_loss = torch.nn.MSELoss()

        self.gan_pgd: PGD_Optimizer = PGD_Optimizer(epsilon=1.0, iteration=500, output=0)

    def reset_parameters(self):
        self.G.apply(weight_init)
        self.D.apply(weight_init)

    def train(self, train_dataloader):
        self.g_optimizer.zero_grad()
        self.d_optimizer.zero_grad()
        for g_iter in range(self.generator_iters):
            # Requires grad, Generator requires_grad = False
            for p in self.D.parameters():
                p.requires_grad = True
                p.data.clamp_(-0.01, 0.01)
            for p in self.G.parameters():
                p.requires_grad = False

            for d_iter in range(self.critic_iter):
                d_loss_fake = 0    # TODO
                d_loss_real = 0    # TODO
                for i, (data, label) in enumerate(train_dataloader):
                    data = torch.tensor(data)
                    train_data = data.to(env['device'])
                    d_loss_real = self.D(train_data).mean()

                    z = torch.randn(train_data.shape[0], self.noise_dim, device=train_data.device)
                    fake_images = self.G(z)
                    d_loss_fake = self.D(fake_images).mean()

                    d_loss = d_loss_fake - d_loss_real
                    d_loss.backward()
                    self.d_optimizer.step()
                    self.d_optimizer.zero_grad()
                print(f'    Discriminator: loss_fake: {d_loss_fake:.5f}, loss_real: {d_loss_real:.5f}')
            for p in self.D.parameters():
                p.requires_grad = False
            for p in self.G.parameters():
                p.requires_grad = True
            g_loss = 0    # TODO
            for i, (data, label) in enumerate(train_dataloader):
                data = torch.tensor(data)
                train_data = data.to(env['device'])
                z = torch.randn(train_data.shape[0], self.noise_dim, device=train_data.device)
                fake_images = self.G(z)
                g_loss = - self.D(fake_images).mean()
                g_loss.backward()
                self.g_optimizer.step()
                self.g_optimizer.zero_grad()
            print(f'Generator iteration: {g_iter:5d} / {self.generator_iters:5d}, g_loss: {g_loss:.5f}')

    def get_encode_value(self, imgs: torch.Tensor):
        """According to the image and Generator, utilize pgd optimization to get the d dimension encoding value.

        Args:
            imgs (torch.FloatTensor): the chosen image to get its encoding value, also considered as the output of Generator.
            noise_dim (int): the dimension of the input in the generator.

        Returns:
            torch.FloatTensor: the synthesized poisoned image.
        """

        def loss_func(X: torch.Tensor):
            loss = self.mse_loss(self.G(X), imgs)
            return loss
        x_1 = torch.randn(len(imgs), self.noise_dim, device=imgs.device)
        x_1, _ = self.gan_pgd.optimize(_input=x_1, loss_fn=loss_func)
        return x_1



#!/usr/bin/env python3

from .imagefolder import ImageFolder
from trojanzoo.utils.param import Module

import torchvision.transforms as transforms
from torchvision.datasets import ImageNet as ImageNet_Official
import numpy as np
import os
import json
from trojanvision import __file__ as root_file
root_dir = os.path.dirname(root_file)


class ImageNet(ImageFolder):

    name = 'imagenet'
    data_shape = [3, 224, 224]
    url = {
        'train': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_train.tar',
        'valid': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_val.tar',
        'test': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_test.tar',
    }
    md5 = {
        'train': '1d675b47d978889d74fa0da5fadfb00e',
        'valid': '29b22e2961454d5413ddabcf34fc5622',
    }

    def __init__(self, norm_par={'mean': [0.485, 0.456, 0.406],
                                 'std': [0.229, 0.224, 0.225], },
                 **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize_folder(self):
        ImageNet_Official(root=self.folder_path, split='train', download=True)
        ImageNet_Official(root=self.folder_path, split='val', download=True)
        os.rename(os.path.join(self.folder_path, 'imagenet', 'val'),
                  os.path.join(self.folder_path, 'imagenet', 'valid'))

    @staticmethod
    def get_transform(mode: str) -> transforms.Compose:
        if mode == 'train':
            transform = transforms.Compose([
                transforms.RandomResizedCrop((224, 224)),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor()])
        else:
            transform = transforms.Compose([
                transforms.Resize((256, 256)),
                transforms.CenterCrop((224, 224)),
                transforms.ToTensor()])
            # BiT transform
            # transform = transforms.Compose([
            #     transforms.Resize((480, 480)),
            #     transforms.ToTensor()])
        return transform

    def initialize_npz(self, mode_list: list[str] = ['train', 'valid'],
                       transform: transforms.Compose = transforms.Compose([transforms.Resize((256, 256)),
                                                                           transforms.Lambda(lambda x: np.array(x))]),
                       **kwargs):
        super().initialize_npz(mode_list=mode_list, transform=transform, **kwargs)


class Sample_ImageNet(ImageNet):

    name: str = 'sample_imagenet'
    num_classes = 10
    url = {}
    org_folder_name = {}

    def initialize_folder(self):
        _dict = Module(self.__dict__)
        _dict.__delattr__('folder_path')
        imagenet = ImageNet(**_dict)
        class_dict: dict = {}
        json_path = os.path.normpath(os.path.join(root_dir, 'data', self.name, 'class_dict.json'))
        with open(json_path, 'r', encoding='utf-8') as f:
            class_dict: dict = json.load(f)
        imagenet.sample(child_name=self.name, class_dict=class_dict)



#!/usr/bin/env python3

from .backdoor_defense import BackdoorDefense
from .adv import *
from .backdoor import *

from trojanvision.configs import Config, config
from trojanvision.datasets import ImageSet
import trojanzoo.defenses
from trojanzoo.defenses import Defense
import argparse
from typing import Union

class_dict = {
    # adversarial Defense
    'advmind': AdvMind,
    'curvature': Curvature,
    # 'grad_train': Grad_Train,
    'adv_train': AdvTrain,

    # backdoor defense
    # model inspection
    'neural_cleanse': NeuralCleanse,
    'tabor': TABOR,
    'abs': ABS,
    'deep_inspect': DeepInspect,

    # input detection
    'strip': STRIP,
    'neo': NEO,

    # training data inspection
    'activation_clustering': ActivationClustering,
    'spectral_signature': SpectralSignature,
    'neuron_inspect': NeuronInspect,

    # general defense
    'fine_pruning': FinePruning,
    'image_transform': ImageTransform,
    'magnet': MagNet,
}


def add_argument(parser: argparse.ArgumentParser, defense_name: str = None, defense: Union[str, Defense] = None,
                 class_dict: dict[str, type[Defense]] = class_dict):
    return trojanzoo.defenses.add_argument(parser=parser, defense_name=defense_name, defense=defense,
                                           class_dict=class_dict)


def create(defense_name: str = None, defense: Union[str, Defense] = None,
           dataset_name: str = None, dataset: Union[str, ImageSet] = None,
           config: Config = config, class_dict: dict[str, type[Defense]] = class_dict, **kwargs):
    return trojanzoo.defenses.create(defense_name=defense_name, defense=defense,
                                     dataset_name=dataset_name, dataset=dataset,
                                     config=config, class_dict=class_dict, **kwargs)



#!/usr/bin/env python3

from .imc import IMC
from trojanvision.marks import Watermark
from trojanzoo.utils.output import prints

import torch
import torch.utils.data
import numpy as np
import math
import random
import os
from typing import Callable


class IMC_Multi(IMC):
    name: str = 'imc_multi'

    def __init__(self, mark_num: int = 5, **kwargs):
        super().__init__(**kwargs)
        self.mark_list: list[tuple[Watermark, int]] = []
        for i in range(mark_num):
            height_offset = random.randint(0, self.mark.data_shape[-2] - self.mark.mark_height)
            width_offset = random.randint(0, self.mark.data_shape[-1] - self.mark.mark_width)
            mark = Watermark(data_shape=self.mark.data_shape, edge_color=self.mark.edge_color, mark_path=self.mark.mark_path,
                             mark_alpha=self.mark.mark_alpha, mark_height=self.mark.mark_height, mark_width=self.mark.mark_width,
                             height_offset=height_offset, width_offset=width_offset,
                             random_init=True, random_pos=False, mark_distributed=self.mark.mark_distributed)
            self.mark_list.append((mark, i))

    def attack(self, epoch: int, save=False, **kwargs):
        self.model._train(epoch, save=save,
                          validate_fn=self.validate_fn, get_data_fn=self.get_train_data,
                          save_fn=self.save, **kwargs)

    # ---------------------- I/O ----------------------------- #

    def save(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        np.save(file_path + '.npy', self.mark_list)
        self.model.save(file_path + '.pth')
        print('attack results saved at: ', file_path)

    def load(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        print('attack results loaded from: ', file_path)
        self.mark_list = np.load(file_path + '.npy', allow_pickle=True)
        self.model.load(file_path + '.pth')

    # ---------------------- Utils ---------------------------- #

    def get_train_data(self, data: tuple[torch.Tensor, torch.Tensor], **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        input_list = [_input]
        label_list = [_label]
        for mark, target_class in self.mark_list:
            decimal, integer = math.modf(self.poison_num)
            integer = int(integer)
            if random.uniform(0, 1) < decimal:
                integer += 1
            if not integer:
                continue
            poison_input = mark.add_mark(_input[:integer])
            poison_label = target_class * torch.ones_like(_label[:integer])
            input_list.append(poison_input)
            label_list.append(poison_label)
        _input = torch.cat(input_list)
        _label = torch.cat(label_list)
        return _input, _label

    def get_poison_data(self, data: tuple[torch.Tensor, torch.Tensor], mark: Watermark = None,
                        target_class: int = 0, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        poison_input = mark.add_mark(_input)
        poison_label = target_class * torch.ones_like(_label)
        return poison_input, poison_label

    def validate_fn(self,
                    get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    loss_fn: Callable[..., torch.Tensor] = None,
                    main_tag: str = 'valid', indent: int = 0, **kwargs) -> tuple[float, float]:
        _, clean_acc = self.model._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                            get_data_fn=None, indent=indent, **kwargs)
        target_acc = 100.0
        for i, (mark, target_class) in enumerate(self.mark_list):
            _, acc = self.model._validate(print_prefix=f'Validate Trigger {i} target {target_class} ', main_tag='',
                                          get_data_fn=self.get_poison_data,
                                          mark=mark, target_class=target_class,
                                          indent=indent, **kwargs)
            target_acc = min(acc, target_acc)
        prints(f'Validate Confidence: {self.validate_confidence():.3f}', indent=indent)
        prints(f'Neuron Jaccard Idx: {self.check_neuron_jaccard():.3f}', indent=indent)
        if self.clean_acc - clean_acc > 3 and self.clean_acc > 40:  # TODO: better not hardcoded
            target_acc = 0.0
        return clean_acc, target_acc



#!/usr/bin/env python3

from .badnet import BadNet
from trojanvision.environ import env
from trojanzoo.utils import to_pil_image, byte2float

import torch
import torch.nn as nn
import numpy as np
import argparse
from PIL import Image


class ReflectionBackdoor(BadNet):
    name: str = 'reflection_backdoor'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--candidate_num', dest='candidate_num', type=int,
                           help='number of candidate images')
        group.add_argument('--selection_num', dest='selection_num', type=int,
                           help='number of adv images')
        group.add_argument('--selection_iter', dest='selection_iter', type=int,
                           help='selection iteration to find optimal reflection images as trigger')
        group.add_argument('--inner_epoch', dest='inner_epoch', type=int,
                           help='retraining epoch during trigger selection')

    def __init__(self, candidate_num: int = 100, selection_num: int = 20, selection_iter: int = 10, inner_epoch: int = 1, **kwargs):
        super().__init__(**kwargs)
        self.param_list['reflection'] = ['candidate_num', 'selection_num', 'selection_iter', 'inner_epoch']
        self.candidate_num: int = candidate_num
        self.selection_iter: int = selection_iter
        self.selection_num: int = selection_num
        self.inner_epoch: int = inner_epoch

        kernel = torch.tensor([[0., 1., 0.],
                               [1., -4., 1.],
                               [0., 1., 0.]], device='cpu')
        self.conv2d = nn.Conv2d(1, 1, 3, bias=False, padding=1)
        self.conv2d.weight = nn.Parameter(kernel.view_as(self.conv2d.weight))

    def attack(self, epoch: int, save=False, validate_interval: int = 10, lr_scheduler=None, **kwargs):
        W = torch.zeros(self.candidate_num)

        loader = self.dataset.get_dataloader(mode='train', batch_size=self.candidate_num, classes=[self.target_class],
                                             shuffle=True, num_workers=0, pin_memory=False)
        candidate_images, _ = next(iter(loader))
        candidate_images = self.conv2d(candidate_images.mean(1, keepdim=True))

        np.random.seed(env['seed'])
        pick_img_ind = np.random.choice(self.candidate_num, self.selection_num, replace=False).tolist()
        adv_images = candidate_images[pick_img_ind]  # (B, C, H, W)

        for current_iter in range(self.selection_iter):
            print(f'Current Iteration : {current_iter}')
            for i in range(len(adv_images)):
                print(f'    adv image idx : {i}')
                self.get_mark(adv_images[i])
                super().attack(self.inner_epoch, indent=8, **kwargs)
                _, target_acc = super().validate_fn(verbose=False)
                W[pick_img_ind[i]] = target_acc
                self.model.load()
            # update W
            if self.selection_num < self.candidate_num:
                other_img_ind = list(set(range(self.candidate_num)) - set(pick_img_ind))
                W[other_img_ind] = W[pick_img_ind].median()
            # re-pick top m reflection images
            pick_img_ind = W.argsort(descending=True).tolist()[:self.selection_num]
            adv_images = candidate_images[pick_img_ind]
        # final training, see performance of best reflection trigger
        self.get_mark(adv_images[0])
        super().attack(epoch, save=save, lr_scheduler=lr_scheduler, **kwargs)

    def get_mark(self, conv_ref_img: torch.Tensor):
        '''
        input is a convolved reflection images, already in same
        shape of any input images, this function will legally reshape
        this ref_img and give to self.mark.mark.
        '''
        org_mark_img: Image.Image = to_pil_image(conv_ref_img)
        org_mark_img = org_mark_img.resize((self.mark.mark_width, self.mark.mark_height), Image.ANTIALIAS)
        self.mark.org_mark = byte2float(org_mark_img)

        self.mark.org_mask, self.mark.org_alpha_mask = self.mark.org_mask_mark(self.mark.org_mark,
                                                                               self.mark.edge_color, self.mark.mark_alpha)
        self.mark.mark, self.mark.mask, self.mark.alpha_mask = self.mark.mask_mark(
            height_offset=self.mark.height_offset, width_offset=self.mark.width_offset)



#!/usr/bin/env python3

__all__ = ['to_tensor', 'to_numpy', 'to_list']

from trojanzoo import __version__

import trojanvision.environ
import trojanvision.datasets
import trojanvision.models
import trojanvision.trainer
import trojanvision.attacks
import trojanvision.defenses

from trojanvision.utils import to_tensor, to_numpy, to_list



#!/usr/bin/env python3

from .badnet import BadNet

from trojanvision.environ import env
from trojanzoo.utils import to_tensor, tanh_func
from trojanzoo.utils import AverageMeter

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import TensorDataset
import argparse

mse_criterion = nn.MSELoss()


class TermStudy(BadNet):
    r"""
    Latent Backdoor Attack is described in detail in the paper `Latent Backdoor`_ by Yuanshun Yao.
    Similar to TrojanNN, Latent Backdoor Attack proposes a method to preprocess the trigger pattern.
    The loss function invloves the distance in feature space.

    .. _Latent Backdoor:
        http://people.cs.uchicago.edu/~huiyingli/publication/fr292-yaoA.pdf
    """
    name: str = 'term_study'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--term', dest='term', type=str)

        group.add_argument('--inner_iter', dest='inner_iter', type=int)
        group.add_argument('--inner_lr', dest='inner_lr', type=float)

        group.add_argument('--class_sample_num', dest='class_sample_num', type=int,
                           help='the number of sampled images per class, defaults to 100')
        group.add_argument('--mse_weight', dest='mse_weight', type=float,
                           help='the weight of mse loss during retraining, defaults to 100')
        group.add_argument('--preprocess_layer', dest='preprocess_layer',
                           help='the chosen feature layer patched by trigger, defaults to \'features\'')
        group.add_argument('--preprocess_epoch', dest='preprocess_epoch', type=int,
                           help='preprocess optimization epoch')
        group.add_argument('--preprocess_lr', dest='preprocess_lr', type=float,
                           help='preprocess learning rate')

    def __init__(self, term='imc', class_sample_num: int = 100, mse_weight=0.5,
                 preprocess_layer: str = 'flatten', preprocess_epoch: int = 100, preprocess_lr: float = 0.1,
                 pgd_iteration: int = 20, pgd_alpha: float = 0.1,
                 **kwargs):
        super().__init__(**kwargs)

        self.param_list['term_study'] = ['term']
        self.term = term
        if term == 'imc':
            self.param_list['imc'] = ['pgd_iteration', 'pgd_alpha']
            self.pgd_iteration: int = pgd_iteration
            self.pgd_alpha: float = pgd_alpha
        elif term == 'latent_backdoor':
            self.param_list['latent_backdoor'] = ['class_sample_num', 'mse_weight',
                                                  'preprocess_layer', 'preprocess_epoch', 'preprocess_lr']
            self.class_sample_num: int = class_sample_num
            self.mse_weight: float = mse_weight

            self.preprocess_layer: str = preprocess_layer
            self.preprocess_epoch: int = preprocess_epoch
            self.preprocess_lr: float = preprocess_lr

            self.avg_target_feats: torch.Tensor = None

    def attack(self, **kwargs):
        if self.term == 'latent_backdoor':
            print('Sample Data')
            data = self.sample_data()
            print('Calculate Average Target Features')
            self.avg_target_feats = self.get_avg_target_feats(data)
            print('Preprocess Mark')
            self.preprocess_mark(data=data)
        elif self.term == 'imc':
            self.optimize_mark()
        print('Retrain')
        return super().attack(**kwargs)

    def sample_data(self) -> dict[str, tuple[torch.Tensor, torch.Tensor]]:
        other_classes = list(range(self.dataset.num_classes))
        other_classes.pop(self.target_class)
        other_x, other_y = [], []
        for _class in other_classes:
            loader = self.dataset.get_dataloader(mode='train', batch_size=self.class_sample_num, classes=[_class],
                                                 shuffle=True, num_workers=0, pin_memory=False)
            _input, _label = next(iter(loader))
            other_x.append(_input)
            other_y.append(_label)
        other_x = torch.cat(other_x)
        other_y = torch.cat(other_y)
        target_loader = self.dataset.get_dataloader(mode='train', batch_size=self.class_sample_num, classes=[self.target_class],
                                                    shuffle=True, num_workers=0, pin_memory=False)
        target_x, target_y = next(iter(target_loader))
        data = {
            'other': (other_x, other_y),
            'target': (target_x, target_y)
        }
        return data
        # other_dataset = torch.utils.data.TensorDataset(other_x, other_y)
        # target_dataset = torch.utils.data.TensorDataset(target_x, target_x)
        # other_loader = self.dataset.get_dataloader(mode='train', dataset=other_dataset, num_workers=0)
        # target_loader = self.dataset.get_dataloader(mode='train', dataset=target_loader, num_workers=0)
        # return other_loader, target_loader

    def get_avg_target_feats(self, data_dict: dict[str, tuple[torch.Tensor, torch.Tensor]]):
        with torch.no_grad():
            if self.dataset.data_shape[1] > 100:
                target_x, target_y = data_dict['target']
                dataset = TensorDataset(target_x, target_y)
                loader = torch.utils.data.DataLoader(dataset=dataset, batch_size=self.dataset.batch_size // max(env['num_gpus'], 1),
                                                     shuffle=True, num_workers=0, pin_memory=False)
                feat_list = []
                for data in loader:
                    target_x, _ = self.model.get_data(data)
                    feat_list.append(self.model.get_layer(target_x, layer_output=self.preprocess_layer).detach().cpu())
                avg_target_feats = torch.cat(feat_list).mean(dim=0)
                avg_target_feats = avg_target_feats.to(target_x.device)
            else:
                target_x, _ = self.model.get_data(data_dict['target'])
                avg_target_feats = self.model.get_layer(target_x, layer_output=self.preprocess_layer).mean(dim=0)
        return avg_target_feats.detach()

    def optimize_mark(self):
        atanh_mark: torch.FloatTensor = torch.randn_like(self.mark.mark) * self.mark.mask
        atanh_mark.requires_grad_()
        self.mark.mark = tanh_func(atanh_mark)
        optimizer = optim.Adam([atanh_mark], lr=self.pgd_alpha)
        optimizer.zero_grad()

        losses = AverageMeter('Loss', ':.4e')
        for _epoch in range(self.pgd_iteration):
            for i, data in enumerate(self.dataset.loader['train']):
                if i > 20:
                    break
                _input, _label = self.model.get_data(data)
                poison_x = self.mark.add_mark(_input)
                loss = self.model.loss(poison_x, self.target_class * torch.ones_like(_label))
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()
                self.mark.mark = tanh_func(atanh_mark)
                losses.update(loss.item(), n=len(_label))
        atanh_mark.requires_grad = False
        self.mark.mark.detach_()

    def preprocess_mark(self, data: dict[str, tuple[torch.Tensor, torch.Tensor]]):
        other_x, _ = data['other']
        other_set = TensorDataset(other_x)
        other_loader = self.dataset.get_dataloader(mode='train', dataset=other_set, num_workers=0)

        atanh_mark = torch.randn_like(self.mark.mark) * self.mark.mask
        atanh_mark.requires_grad_()
        self.mark.mark = tanh_func(atanh_mark)
        optimizer = optim.Adam([atanh_mark], lr=self.preprocess_lr)
        optimizer.zero_grad()

        losses = AverageMeter('Loss', ':.4e')
        for _epoch in range(self.preprocess_epoch):
            # epoch_start = time.perf_counter()
            loader = other_loader
            # if env['tqdm']:
            #     loader = tqdm(loader)
            for (batch_x, ) in loader:
                poison_x = self.mark.add_mark(to_tensor(batch_x))
                loss = self.loss_mse(poison_x)
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()
                self.mark.mark = tanh_func(atanh_mark)
                losses.update(loss.item(), n=len(batch_x))
            # epoch_time = str(datetime.timedelta(seconds=int(
            #     time.perf_counter() - epoch_start)))
            # pre_str = '{blue_light}Epoch: {0}{reset}'.format(
            #     output_iter(_epoch + 1, self.preprocess_epoch), **ansi).ljust(64 if env['color'] else 35)
            # _str = ' '.join([
            #     f'Loss: {losses.avg:.4f},'.ljust(20),
            #     f'Time: {epoch_time},'.ljust(20),
            # ])
            # prints(pre_str, _str, prefix='{upline}{clear_line}'.format(**ansi) if env['tqdm'] else '', indent=4)
        atanh_mark.requires_grad = False
        self.mark.mark.detach_()

    def loss_mse(self, poison_x: torch.Tensor) -> torch.Tensor:
        other_feats = self.model.get_layer(poison_x, layer_output=self.preprocess_layer)
        return mse_criterion(other_feats, self.avg_target_feats)



#!/usr/bin/env python3

# todo: Need a better name

import trojanzoo.optim

from trojanzoo.utils import atan_func, tanh_func
from trojanzoo.utils.output import prints

import torch
import torch.optim
from torch.optim import Optimizer
from torch.optim.lr_scheduler import StepLR
from collections.abc import Callable
from typing import Any, Union


class Uname(trojanzoo.optim.Optimizer):
    r"""This class transforms input (tanh, atan or sigmoid) and then apply standard torch.optim.Optimizer
    """

    name: str = 'uname'

    def __init__(self, OptimType: Union[str, type[Optimizer]], optim_kwargs: dict[str, Any] = {},
                 lr_scheduler: bool = False, lr_step_size: int = 50,
                 input_transform: Union[str, Callable] = lambda x: x, **kwargs):    # TODO: Callable[[torch.Tensor], torch.Tensor]
        super().__init__(**kwargs)
        self.param_list['uname'] = ['OptimType', 'optim_kwargs', 'lr_scheduler', 'lr_step_size', 'input_transform']
        if isinstance(OptimType, str):
            OptimType = getattr(torch.optim, OptimType)
        self.OptimType: type[Optimizer] = OptimType
        self.optim_kwargs: dict = optim_kwargs
        self.lr_scheduler: bool = lr_scheduler
        self.lr_step_size: int = lr_step_size
        self.input_transform: Callable[[torch.Tensor], torch.Tensor] = input_transform

    def optimize(self, unbound_params: list[torch.Tensor],
                 iteration: int = None, loss_fn: Callable[[torch.Tensor], torch.Tensor] = None,
                 output: Union[int, list[str]] = None, **kwargs) -> tuple[list[torch.Tensor], int]:
        # ------------------------------ Parameter Initialization ---------------------------------- #
        iteration = iteration if iteration is not None else self.iteration
        loss_fn = loss_fn if loss_fn is not None else self.loss_fn
        output = self.get_output(output)
        if isinstance(unbound_params, torch.Tensor):
            unbound_params = [unbound_params]

        # ----------------------------------------------------------------------------------------- #
        real_params: list[torch.Tensor] = []
        for param in unbound_params:
            param.requires_grad_()
            real_params.append(self.transform_func(param))
        if 'start' in output:
            self.output_info(real_params=real_params, mode='start', loss_fn=loss_fn, **kwargs)
        if iteration == 0:
            return real_params, None
        optimizer = self.OptimType(parameters=unbound_params, **self.optim_kwargs)
        lr_scheduler = StepLR(optimizer, lr_step_size=self.lr_step_size) if self.lr_scheduler else None
        optimizer.zero_grad()

        # ----------------------------------------------------------------------------------------- #

        for _iter in range(iteration):
            with torch.no_grad():
                loss_value = float(loss_fn(*real_params))
            if self.early_stop_check(loss_value):
                for param in unbound_params:
                    param.requires_grad_(False)
                for param in real_params:
                    param.detach_()
                if 'end' in output:
                    self.output_info(real_params=real_params, mode='end', loss_fn=loss_fn, **kwargs)
                return real_params, _iter + 1
            loss = loss_fn(*real_params)
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
            real_params = []
            for param in unbound_params:
                real_params.append(self.transform_func(param))
            if lr_scheduler:
                lr_scheduler.step()
            if 'middle' in output:
                self.output_info(real_params=real_params, mode='middle',
                                 _iter=_iter, iteration=iteration, loss_fn=loss_fn, **kwargs)
        for param in unbound_params:
            param.requires_grad_(False)
        for param in real_params:
            param.detach_()
        if 'end' in output:
            self.output_info(real_params=real_params, mode='end', loss_fn=loss_fn, **kwargs)
        return real_params, None

    def transform_func(self, x: torch.Tensor) -> torch.Tensor:
        if isinstance(self.input_transform, str):
            if self.input_transform == 'tanh':
                return tanh_func(x)
            elif self.input_transform in ['atan', 'arctan']:
                return atan_func(x)
            elif self.input_transform in ['sigmoid', 'logistic']:
                return torch.sigmoid(x)
            else:
                raise NotImplementedError(self.input_transform)
        # assert callable(self.input_transform)
        return self.input_transform(x)

    def output_info(self, real_params: torch.Tensor, loss_fn: Callable[..., torch.Tensor] = None, **kwargs):
        super().output_info(**kwargs)
        with torch.no_grad():
            loss = float(loss_fn(*real_params))
        prints(f'loss: {loss:.5f}', indent=self.indent)



#!/usr/bin/env python3

from trojanvision.configs import Config, config
import trojanzoo.trainer

from typing import TYPE_CHECKING
from trojanvision.datasets import ImageSet    # TODO: python 3.10
from trojanvision.models import ImageModel
import argparse
if TYPE_CHECKING:
    pass


class Trainer(trojanzoo.trainer.Trainer):
    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
        super().add_argument(group)
        group.add_argument('--adv_train', dest='adv_train', action='store_true',
                           help='enable adversarial training.')
        group.add_argument('--adv_train_alpha', dest='adv_train_alpha', type=float,
                           help='adversarial training PGD alpha, defaults to 2/255.')
        group.add_argument('--adv_train_epsilon', dest='adv_train_epsilon', type=float,
                           help='adversarial training PGD epsilon, defaults to 8/255.')
        group.add_argument('--adv_train_iter', dest='adv_train_iter', type=int,
                           help='adversarial training PGD iteration, defaults to 7.')


def add_argument(parser: argparse.ArgumentParser, ClassType: type[Trainer] = Trainer) -> argparse._ArgumentGroup:
    return trojanzoo.trainer.add_argument(parser=parser, ClassType=ClassType)


def create(dataset_name: str = None, dataset: ImageSet = None, model: ImageModel = None,
           config: Config = config, ClassType=Trainer, **kwargs):
    return trojanzoo.trainer.create(dataset_name=dataset_name, dataset=dataset,
                                    ClassType=ClassType,
                                    model=model, config=config, **kwargs)



#!/usr/bin/env python3

from .poison_basic import PoisonBasic
from trojanvision.attacks import PGD
from trojanvision.defenses.adv.curvature import Curvature
from trojanvision.optim import PGD as PGD_Optimizer
from trojanvision.models import MagNet
from trojanzoo.utils import to_list

import torch
import numpy as np
import os
from scipy.stats import ks_2samp
import argparse


class IMC_Poison(PoisonBasic):

    name: str = 'imc_poison'

    # TODO: change PGD to Uname.optimizer
    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--pgd_alpha', dest='pgd_alpha', type=float)
        group.add_argument('--pgd_epsilon', dest='pgd_epsilon', type=float)
        group.add_argument('--pgd_iteration', dest='pgd_iteration', type=int)
        group.add_argument('--stop_conf', dest='stop_conf', type=float)

        group.add_argument('--magnet', dest='magnet', action='store_true')
        group.add_argument('--randomized_smooth', dest='randomized_smooth', action='store_true')
        group.add_argument('--curvature', dest='curvature', action='store_true')

    def __init__(self, pgd_alpha: float = 1.0, pgd_epsilon: float = 8.0, pgd_iteration: int = 8,
                 stop_conf: float = 0.9,
                 magnet: bool = False, randomized_smooth: bool = False, curvature: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.param_list['pgd'] = ['pgd_alpha', 'pgd_epsilon', 'pgd_iteration']
        self.pgd_alpha: float = pgd_alpha
        self.pgd_epsilon: float = pgd_epsilon
        self.pgd_iteration: int = pgd_iteration
        self.pgd = PGD_Optimizer(alpha=self.pgd_alpha / 255, epsilon=self.pgd_epsilon / 255,
                                 iteration=self.pgd_iteration)
        self.stop_conf: float = stop_conf
        if magnet:
            self.magnet: MagNet = MagNet(dataset=self.dataset, pretrain=True)
        self.randomized_smooth: bool = randomized_smooth
        if curvature:
            self.curvature: Curvature = Curvature(model=self.model)

    def attack(self, epoch: int, **kwargs):
        # model._validate()
        total = 0
        target_conf_list = []
        target_acc_list = []
        clean_acc_list = []
        pgd_norm_list = []
        alpha = 1.0 / 255
        epsilon = 8.0 / 255
        if self.dataset.name in ['cifar10', 'gtsrb', 'isic2018']:
            alpha = 1.0 / 255
            epsilon = 8.0 / 255
        if self.dataset.name in ['sample_imagenet', 'sample_vggface2']:
            alpha = 0.25 / 255
            epsilon = 2.0 / 255
        pgd_checker = PGD(alpha=alpha, epsilon=epsilon, iteration=8,
                          dataset=self.dataset, model=self.model, target_idx=self.target_idx, stop_threshold=0.95)
        easy = 0
        difficult = 0
        normal = 0
        loader = self.dataset.get_dataloader(mode='valid', batch_size=self.dataset.test_batch_size)
        if 'curvature' in self.__dict__.keys():
            benign_curvature = self.curvature.benign_measure()
            tgt_curvature_list = []
            org_curvature_list = []
        if self.randomized_smooth:
            org_conf_list = []
            tgt_conf_list = []
        if 'magnet' in self.__dict__.keys():
            org_magnet_list = []
            tgt_magnet_list = []
        for data in loader:
            print(easy, normal, difficult)
            if normal >= 100:
                break
            self.model.load()
            _input, _label = self.model.remove_misclassify(data)
            if len(_label) == 0:
                continue
            target_label = self.model.generate_target(_input, idx=self.target_idx)
            self.temp_input = _input
            self.temp_label = target_label
            _, _iter = pgd_checker.craft_example(_input)
            if _iter is None:
                difficult += 1
                continue
            if _iter < 4:
                easy += 1
                continue
            normal += 1
            target_conf, target_acc, clean_acc = self.validate_fn()
            noise = torch.zeros_like(_input)
            poison_input = self.craft_example(_input=_input, _label=target_label, epoch=epoch, noise=noise, **kwargs)
            pgd_norm = float(noise.norm(p=float('inf')))
            target_conf, target_acc, clean_acc = self.validate_fn()
            target_conf_list.append(target_conf)
            target_acc_list.append(target_acc)
            clean_acc_list.append(max(self.clean_acc - clean_acc, 0.0))
            pgd_norm_list.append(pgd_norm)
            print(f'[{total+1} / 100]\n'
                  f'target confidence: {np.mean(target_conf_list)}({np.std(target_conf_list)})\n'
                  f'target accuracy: {np.mean(target_acc_list)}({np.std(target_acc_list)})\n'
                  f'clean accuracy Drop: {np.mean(clean_acc_list)}({np.std(clean_acc_list)})\n'
                  f'PGD Norm: {np.mean(pgd_norm_list)}({np.std(pgd_norm_list)})\n\n\n')
            org_conf = self.model.get_target_prob(_input=poison_input, target=_label)
            tgt_conf = self.model.get_target_prob(_input=poison_input, target=target_label)
            if 'curvature' in self.__dict__.keys():
                org_curvature_list.extend(to_list(self.curvature.measure(poison_input, _label)))    # type: ignore
                tgt_curvature_list.extend(to_list(self.curvature.measure(poison_input, target_label)))    # type: ignore
                print('Curvature:')
                print(f'    org_curvature: {ks_2samp(org_curvature_list, benign_curvature)}')    # type: ignore
                print(f'    tgt_curvature: {ks_2samp(tgt_curvature_list, benign_curvature)}')    # type: ignore
                print()
            if self.randomized_smooth:
                org_new = self.model.get_target_prob(_input=poison_input, target=_label, randomized_smooth=True)
                tgt_new = self.model.get_target_prob(_input=poison_input, target=target_label, randomized_smooth=True)
                org_increase = (org_new - org_conf).clamp(min=0.0)
                tgt_decrease = (tgt_new - tgt_conf).clamp(min=0.0)
                org_conf_list.extend(to_list(org_increase))    # type: ignore
                tgt_conf_list.extend(to_list(tgt_decrease))    # type: ignore
                print('Randomized Smooth:')
                print(f'    org_confidence: {np.mean(org_conf_list)}')    # type: ignore
                print(f'    tgt_confidence: {np.mean(tgt_conf_list)}')    # type: ignore
                print()
            if 'magnet' in self.__dict__.keys():
                poison_input = self.magnet(poison_input)
                org_new = self.model.get_target_prob(_input=poison_input, target=_label)
                tgt_new = self.model.get_target_prob(_input=poison_input, target=target_label)
                org_increase = (org_new - org_conf).clamp(min=0.0)
                tgt_decrease = (tgt_conf - tgt_new).clamp(min=0.0)
                org_magnet_list.extend(to_list(org_increase))    # type: ignore
                tgt_magnet_list.extend(to_list(tgt_decrease))    # type: ignore
                print('MagNet:')
                print(f'    org_confidence: {np.mean(org_magnet_list)}')    # type: ignore
                print(f'    tgt_confidence: {np.mean(tgt_magnet_list)}')    # type: ignore
                print()
            total += 1

    def craft_example(self, _input: torch.Tensor, _label: torch.Tensor, noise: torch.Tensor = None, save=False, **kwargs):
        if noise is None:
            noise = torch.zeros_like(_input)
        poison_input = None
        for _iter in range(self.pgd_iteration):
            target_conf, target_acc = self.validate_target(indent=4, verbose=False)
            if target_conf > self.stop_conf:
                break
            poison_input, _ = self.pgd.optimize(_input, noise=noise, loss_fn=self.loss_pgd, iteration=1)
            self.temp_input = poison_input
            target_conf, target_acc = self.validate_target(indent=4, verbose=False)
            if target_conf > self.stop_conf:
                break
            self._train(_input=poison_input, _label=_label, **kwargs)
        target_conf, target_acc = self.validate_target(indent=4, verbose=False)
        return poison_input

    def save(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.model.save(file_path + '.pth')
        print('attack results saved at: ', file_path)

    def get_filename(self, **kwargs):
        return self.model.name

    def loss_pgd(self, x: torch.Tensor) -> torch.Tensor:
        return self.model.loss(x, self.temp_label)
